# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2009-2013 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2010-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""File management for brushes and brush groups."""

## Imports

from __future__ import division, print_function

from itertools import chain
import os
import zipfile
from os.path import basename
from warnings import warn
import logging
import shutil
import uuid
import contextlib

from lib.gettext import gettext as _
from lib.gettext import C_
from lib.helpers import utf8

from lib.gibindings import Gtk
from lib.gibindings import GdkPixbuf

from . import dialogs
from lib.brush import BrushInfo
from lib.observable import event
import lib.pixbuf
from . import drawutils
import gui.mode
import lib.config
from lib.pycompat import unicode
from lib.pycompat import xrange
from lib.pycompat import PY3

if PY3:
    import urllib.parse
else:
    import urllib


## Public module constants

PREVIEW_W = 128   #: Width of brush preview images
PREVIEW_H = 128   #: Height of brush preview images

FOUND_BRUSHES_GROUP = 'lost&found'   #: Orphaned brushes found at startup
DELETED_BRUSH_GROUP = 'deleted'   #: Orphaned brushes go here after group del
FAVORITES_BRUSH_GROUP = u'favorites'  #: User's favourites
NEW_BRUSH_GROUP = 'new'  #: Home for newly created brushes

## Internal module constants

_DEFAULT_STARTUP_GROUP = 'set#2'  # Suggestion only (FIXME: no effect?)
_DEFAULT_BRUSH = 'Dieterle/Fan#1'  # TODO: phase out and use heuristics?
_DEFAULT_ERASER = 'deevad/kneaded_eraser_large'  # TODO: -----------"---------
_DEVBRUSH_NAME_PREFIX = "devbrush_"
_BRUSH_HISTORY_NAME_PREFIX = "history_"
_BRUSH_HISTORY_SIZE = 5
_NUM_BRUSHKEYS = 10

_BRUSHPACK_README = "readme.txt"
_BRUSHPACK_ORDERCONF = "order.conf"

_DEVICE_NAME_NAMESPACE = uuid.UUID('169eaf8a-554e-45b8-8295-fc09b10031cc')

_TEST_BRUSHPACK_PY27 = u"tests/brushpacks/saved-with-py2.7.zip"

logger = logging.getLogger(__name__)


## Helper functions

def _device_name_uuid(device_name):
    """Return UUID5 string for a given device name

    >>> result = _device_name_uuid(u'Wacom Intuos5 touch S Pen stylus')
    >>> result == u'e97830e9-f9f9-50a5-8fff-68bead1a7021'
    True
    >>> type(result) == type(u'')
    True

    """
    if not PY3:
        device_name = utf8(unicode(device_name))
    return unicode(uuid.uuid5(_DEVICE_NAME_NAMESPACE, device_name))


def _quote_device_name(device_name):
    """Converts a device name to something safely storable on the disk

    Quotes an arbitrary device name for use as the basename of a
    device-specific brush.

    >>> result = _quote_device_name(u'Heavy Metal Umlaut D\u00ebvice')
    >>> result == 'Heavy+Metal+Umlaut+D%C3%ABvice'
    True
    >>> type(result) == type(u'')
    True
    >>> result = _quote_device_name(u'unsafe/device\\\\name')
    >>> result == 'unsafe%2Fdevice%5Cname'
    True
    >>> type(result) == type(u'')
    True

    Hopefully this is OK for Windows, UNIX and Mac OS X names.
    """
    device_name = unicode(device_name)
    if PY3:
        quoted = urllib.parse.quote_plus(
            device_name, safe='',
            encoding="utf-8",
        )
    else:
        u8bytes = device_name.encode("utf-8")
        quoted = urllib.quote_plus(u8bytes, safe='')
    return unicode(quoted)


def translate_group_name(name):
    """Translates a group name from a disk name to a display name."""
    d = {FOUND_BRUSHES_GROUP: _('Lost & Found'),
         DELETED_BRUSH_GROUP: _('Deleted'),
         FAVORITES_BRUSH_GROUP: _('Favorites'),
         'ink': _('Ink'),
         'classic': _('Classic'),
         'set#1': _('Set#1'),
         'set#2': _('Set#2'),
         'set#3': _('Set#3'),
         'set#4': _('Set#4'),
         'set#5': _('Set#5'),
         'experimental': _('Experimental'),
         'new': _('New'),
         }
    return d.get(name, name)


def _parse_order_conf(file_content):
    """Parse order.conf file data.

    :param bytes file_content: data from an order.conf (encoded UTF-8)
    :returns: a group dict

    The returned dict is of the form "{u'group1' : [u'brush1',
    u'brush2'], u'group2' : [u'brush3']}".

    """
    groups = {}
    try:
        file_content = file_content.decode("utf-8")
    except UnicodeDecodeError:
        # This handles order.conf files saved with the wrong encoding
        # on Windows (encoding was previously not explicitly specified).
        logger.warning("order.conf file not encoded with utf-8")
        file_content = file_content.decode('latin-1')
    curr_group = FOUND_BRUSHES_GROUP
    lines = file_content.replace(u'\r', u'\n').split(u'\n')
    for line in lines:
        name = line.strip()
        if name.startswith(u'#') or not name:
            continue
        if name.startswith(u'Group: '):
            curr_group = name[7:]
            if curr_group not in groups:
                groups[curr_group] = []
            continue
        groups.setdefault(curr_group, [])
        if name in groups[curr_group]:
            logger.warning(
                '%r: brush appears twice in the same group, ignored',
                name,
            )
            continue
        groups[curr_group].append(name)
    return groups


## Class definitions


class BrushManager (object):
    """Brush manager, responsible for groups of brushes."""

    ## Initialization

    def __init__(self, stock_brushpath, user_brushpath, app=None):
        """Initialize, with paths and a ref to the main app.

        :param unicode|str stock_brushpath: MyPaint install's RO brushes.
        :param unicode|str user_brushpath: User-writable brush library.
        :param gui.application.Application app: Main app (use None for test).

        The user_brushpath folder will be created if it does not yet exist.

        >>> from tempfile import mkdtemp
        >>> from shutil import rmtree
        >>> tmpdir = mkdtemp(u".brushes")
        >>> bm = BrushManager(lib.config.mypaint_brushdir, tmpdir, app=None)
        >>> len(bm.groups) > 0
        True
        >>> all([isinstance(k, unicode) for k in bm.groups.keys()])
        True
        >>> all([isinstance(v, list) for v in bm.groups.values()])
        True
        >>> rmtree(tmpdir)

        """
        super(BrushManager, self).__init__()

        # Default pigment setting when not specified by the brush
        self.pigment_by_default = None

        self.stock_brushpath = stock_brushpath
        self.user_brushpath = user_brushpath
        self.app = app

        #: The selected brush, as a ManagedBrush. Its settings are
        #: automatically reflected into the working brush engine brush when
        #: it changes.
        self.selected_brush = None

        self.groups = {}  #: Lists of ManagedBrushes, keyed by group name
        self.contexts = []  # Brush keys, indexed by keycap digit number
        self._brush_by_device = {}  # Device name to brush mapping.

        #: Slot used elsewhere for storing the ManagedBrush corresponding to
        #: the most recently saved or restored "context", a.k.a. brush key.
        self.selected_context = None

        if not os.path.isdir(self.user_brushpath):
            os.mkdir(self.user_brushpath)
        self._init_groups()

        # Brush order saving when that changes.
        self.brushes_changed += self._brushes_modified_cb

        # Update the history at the end of each definite input stroke.
        if app is not None:
            app.doc.input_stroke_ended += self._input_stroke_ended_cb

        # Make sure the user always gets a brush tool when they pick a brush
        # preset.
        self.brush_selected += self._brush_selected_cb

    @classmethod
    @contextlib.contextmanager
    def _mock(cls):
        """Context-managed mock BrushManager object for tests.

        Brushes are imported from the shipped brushes subfolder,
        and the user temp area is a temporary directory that's
        cleaned up by the context manager.

        Body yields (BrushManager_instance, tmp_dir_path).

        Please ensure that there are no open files in the tmpdir after
        use to that it can be rmtree()d. On Windows, that means closing
        any zipfile.Zipfile()s you open, even for read.

        """
        from tempfile import mkdtemp
        from shutil import rmtree

        dist_brushes = lib.config.mypaint_brushdir
        tmp_user_brushes = mkdtemp(suffix=u"_brushes")
        try:
            bm = cls(dist_brushes, tmp_user_brushes, app=None)
            yield (bm, tmp_user_brushes)
        finally:
            rmtree(tmp_user_brushes)

    def _load_brush(self, brush_cache, name, **kwargs):
        """Load a ManagedBrush from disk by name, via a cache."""
        if name not in brush_cache:
            b = ManagedBrush(self, name, persistent=True, **kwargs)
            brush_cache[name] = b
        return brush_cache[name]

    def _load_ordered_groups(self, brush_cache, filename):
        try:
            return self._load_ordered_groups_inner(brush_cache, filename)
        except Exception:
            logger.exception("Failed to load groups from %s" % filename)
            return {}

    def _load_ordered_groups_inner(self, brush_cache, filename):
        """Load a groups dict from an order.conf file."""
        groups = {}
        if os.path.exists(filename):
            with open(filename, "rb") as fp:
                groups = _parse_order_conf(fp.read())
            # replace brush names with ManagedBrush instances
            for group, names in list(groups.items()):
                brushes = []
                for name in names:
                    try:
                        b = self._load_brush(brush_cache, name)
                    except IOError as e:
                        logger.warn('%r: %r (removed from group)', name, e)
                        continue
                    brushes.append(b)
                groups[group] = brushes
        return groups

    def _init_ordered_groups(self, brush_cache):
        """Initialize the ordered subset of available brush groups.

        The ordered subset consists of those brushes which are listed in
        the stock and user brush directories' `order.conf` files.  This
        method safely merges upstream changes into the user's ordering.

        """
        join = os.path.join
        base_order_conf = join(self.user_brushpath, 'order_default.conf')
        our_order_conf = join(self.user_brushpath, 'order.conf')
        their_order_conf = join(self.stock_brushpath, 'order.conf')

        # Three-way-merge of brush groups (for upgrading)
        base = self._load_ordered_groups(brush_cache, base_order_conf)
        our = self._load_ordered_groups(brush_cache, our_order_conf)
        their = self._load_ordered_groups(brush_cache, their_order_conf)

        if not our:
            # order.conf missing, restore stock order even
            # if order_default.conf exists
            base = {}

        if base == their:
            self.groups = our
        else:
            logger.info('Merging upstream brush changes into your collection.')
            groups = set(base).union(our).union(their)
            for group in groups:
                # treat the non-existing groups as if empty
                base_brushes = base.setdefault(group, [])
                our_brushes = our.setdefault(group, [])
                their_brushes = their.setdefault(group, [])
                # add new brushes
                insert_index = 0
                for b in their_brushes:
                    if b in our_brushes:
                        insert_index = our_brushes.index(b) + 1
                    else:
                        if b not in base_brushes:
                            our_brushes.insert(insert_index, b)
                            insert_index += 1
                # remove deleted brushes
                for b in base_brushes:
                    if b not in their_brushes and b in our_brushes:
                        our_brushes.remove(b)
                # remove empty groups (except for the favorites)
                if not our_brushes and group != FAVORITES_BRUSH_GROUP:
                    del our[group]
            # finish
            self.groups = our
            self.save_brushorder()
            shutil.copy(their_order_conf, base_order_conf)

    def _list_brushes(self, path):
        """Recursively list the brushes within a directory.

        Return a list of brush names relative to path, using slashes
        for subdirectories on all platforms.

        """
        path += '/'
        result = []
        assert isinstance(path, unicode)  # make sure we get unicode filenames
        for name in os.listdir(path):
            assert isinstance(name, unicode)
            if name.endswith('.myb'):
                result.append(name[:-4])
            elif os.path.isdir(path + name):
                for name2 in self._list_brushes(path + name):
                    result.append(name + '/' + name2)
        return result

    def _init_unordered_groups(self, brush_cache):
        """Initialize the unordered subset of available brushes+groups.

        The unordered subset consists of all brushes that are not listed
        in an `order.conf` file. It includes brushkey brushes,
        per-device brushes, brushes in the painting history.

        This method trawls the stock and user brush directories for
        brushes which aren't listed in in an existing group, and adds
        them to the Lost & Found group, creating it if necessary. It
        should therefore be called after `_init_ordered_groups()`.

        """
        listbrushes = self._list_brushes
        for name in (listbrushes(self.stock_brushpath)
                     + listbrushes(self.user_brushpath)):
            if name.startswith(_DEVBRUSH_NAME_PREFIX):
                # Device brushes are lazy-loaded in fetch_brush_for_device()
                continue

            try:
                b = self._load_brush(brush_cache, name)
            except IOError as e:
                logger.warn("%r: %r (ignored)", name, e)
                continue
            if name.startswith('context'):
                i = int(name[-2:])
                self.contexts[i] = b
            elif name.startswith(_BRUSH_HISTORY_NAME_PREFIX):
                i_str = name.replace(_BRUSH_HISTORY_NAME_PREFIX, '')
                i = int(i_str)
                if 0 <= i < _BRUSH_HISTORY_SIZE:
                    self.history[i] = b
                else:
                    logger.warning(
                        "Brush history item %s "
                        "(entry %d): index outside of history range (0-%d)!",
                        name, i,
                        _BRUSH_HISTORY_SIZE - 1
                    )
            else:
                if not self.is_in_brushlist(b):
                    logger.info("Unassigned brush %r: assigning to %r",
                                name, FOUND_BRUSHES_GROUP)
                    brushes = self.groups.setdefault(FOUND_BRUSHES_GROUP, [])
                    brushes.insert(0, b)

    def _init_default_brushkeys_and_history(self):
        """Assign sensible defaults for brushkeys and history.

        Operates by filling in the gaps after `_init_unordered_groups()`
        has had a chance to populate the two lists.

        """

        # Try the default startup group first.
        default_group = self.groups.get(_DEFAULT_STARTUP_GROUP, None)

        # Otherwise, use the biggest group to minimise the chance
        # of repetition.
        if default_group is None:
            groups_by_len = sorted((len(g), n, g)
                                   for n, g in self.groups.items())
            _len, _name, default_group = groups_by_len[-1]

        # Populate blank entries.
        for i in xrange(_NUM_BRUSHKEYS):
            if self.contexts[i] is None:
                idx = (i + 9) % 10  # keyboard order
                c_name = unicode('context%02d') % i
                c = ManagedBrush(self, name=c_name, persistent=False)
                group_idx = idx % len(default_group)
                b = default_group[group_idx]
                b.clone_into(c, c_name)
                self.contexts[i] = c
        for i in xrange(_BRUSH_HISTORY_SIZE):
            if self.history[i] is None:
                h_name = unicode('%s%d') % (_BRUSH_HISTORY_NAME_PREFIX, i)
                h = ManagedBrush(self, name=h_name, persistent=False)
                group_i = i % len(default_group)
                b = default_group[group_i]
                b.clone_into(h, h_name)
                self.history[i] = h

    def _init_groups(self):
        """Initialize brush groups, loading them from disk."""

        self.contexts = [None for i in xrange(_NUM_BRUSHKEYS)]
        self.history = [None for i in xrange(_BRUSH_HISTORY_SIZE)]

        brush_cache = {}
        self._init_ordered_groups(brush_cache)
        self._init_unordered_groups(brush_cache)
        self._init_default_brushkeys_and_history()

        # clean up legacy stuff
        fn = os.path.join(self.user_brushpath, 'deleted.conf')
        if os.path.exists(fn):
            os.remove(fn)

    ## Observable events

    @event
    def brushes_changed(self, brushes):
        """Event: brushes changed (within their groups).

        Each observer is called with the following args:

        :param self: this BrushManager object
        :param brushes: Affected brushes
        :type brushes: list of ManagedBrushes

        This event is used to notify about brush ordering changes or brushes
        being moved between groups.
        """

    @event
    def groups_changed(self):
        """Event: brush groups changed (deleted, renamed, created)

        Observer callbacks are invoked with no args (other than a ref to the
        brushgroup).  This is used when the "set" of groups change, e.g. when a
        group is renamed, deleted, or created.  It's invoked when self.groups
        changes.
        """

    @event
    def brush_selected(self, brush, info):
        """Event: a different brush was selected.

        Observer callbacks are invoked with the newly selected ManagedBrush and
        its corresponding BrushInfo.
        """

    ## Initial and default brushes

    def select_initial_brush(self):
        """Select the initial brush using saved app preferences.
        """
        initial_brush = None
        # If we recorded which devbrush was last in use, restore it and assume
        # that most of the time the user will continue to work with the same
        # brush and its settings.
        app = self.app
        if app is not None:
            prefs = app.preferences
            last_used_devbrush = prefs.get('devbrush.last_used')
            initial_brush = self.fetch_brush_for_device(last_used_devbrush)
            # Otherwise, initialise from the old selected_brush setting
            if initial_brush is None:
                last_active_name = prefs.get('brushmanager.selected_brush')
                if last_active_name is not None:
                    initial_brush = self.get_brush_by_name(last_active_name)
        # Fallback
        if initial_brush is None:
            initial_brush = self.get_default_brush()
        self.select_brush(initial_brush)

    def _get_matching_brush(self, name=None, keywords=None,
                            favored_group=_DEFAULT_STARTUP_GROUP,
                            fallback_eraser=0.0):
        """Gets a brush robustly by name, by partial name, or a default.

        If a brush named `name` exists, use that. Otherwise search though all
        groups, `favored_group` first, for brushes with any of `keywords`
        in their name. If that fails, construct a new default brush and use
        a given value for its 'eraser' property.
        """
        if name is not None:
            brush = self.get_brush_by_name(name)
            if brush is not None:
                return brush
        if keywords is not None:
            group_names = sorted(self.groups.keys())
            if favored_group in self.groups:
                group_names.remove(favored_group)
                group_names.insert(0, favored_group)
            for group_name in group_names:
                for brush in self.groups[group_name]:
                    for keyword in keywords:
                        if keyword in brush.name:
                            return brush
        # Fallback
        name = 'fallback-default'
        if fallback_eraser != 0.0:
            name += '-eraser'
        brush = ManagedBrush(self, name)
        brush.brushinfo.set_base_value("eraser", fallback_eraser)
        return brush

    def get_default_brush(self):
        """Returns a suitable default drawing brush."""
        drawing = ["pencil", "charcoal", "sketch"]
        return self._get_matching_brush(name=_DEFAULT_BRUSH, keywords=drawing)

    def get_default_eraser(self):
        """Returns a suitable default eraser brush."""
        erasing = ["eraser", "kneaded", "smudge"]
        return self._get_matching_brush(name=_DEFAULT_ERASER, keywords=erasing,
                                        fallback_eraser=1.0)

    def set_pigment_by_default(self, pigment_by_default):
        """Change the default pigment setting to on/off

        This updates loaded managed brushes as well, if they
        do not have the pigment setting set explicitly.
        """
        if self.pigment_by_default != pigment_by_default:
            msg = "Switching default pigment setting to {state}"
            logger.info(msg.format(
                state="On" if pigment_by_default else "Off"))
            self.pigment_by_default = pigment_by_default
            self._reset_pigment_setting()

    def default_pigment_setting(self, setting_info):
        """Pigment (paint_mode) setting override
        """
        if self.pigment_by_default:
            return setting_info.default
        else:
            return setting_info.min

    def _reset_pigment_setting(self):
        appbrush = ()
        if self.app:
            appbrush = (self.app.brush,)
        # Reset the pigment setting for any cached brushes
        # that may have been loaded with the other default - this
        # will not affect brushes that have the pigment setting
        # defined explicitly.
        to_reset = chain(
            # Alter both the current working brush and the selected brush
            appbrush,
            (self.selected_brush.get_brushinfo(),),
            # Also the brush history
            [mb.get_brushinfo() for mb in self.history
             if mb is not None and mb.loaded()],
            # And any other loaded brush
            [mb.get_brushinfo()
             for v in self.groups.values()
             for mb in v if mb.loaded()]
        )
        for bi in to_reset:
            bi.reset_if_undefined('paint_mode')


    ## Brushpack import and export

    def import_brushpack(self, path, window=None):
        """Import a brushpack from a zipfile, with confirmation dialogs.

        :param path: Brush pack zipfile path
        :type path: str
        :param window: Parent window, for dialogs to set.
        :type window: GtkWindow or None
        :returns: Set of imported group names
        :rtype: set

        Confirmation dialogs are only shown if "window" is a suitable
        toplevel to attach the dialogs to.

        >>> with BrushManager._mock() as (bm, tmpdir):
        ...     imp = bm.import_brushpack(_TEST_BRUSHPACK_PY27, window=None)
        ...     py27_g = bm.groups.get(list(imp)[0])
        >>> py27_g[0]  # doctest: +ELLIPSIS
        <ManagedBrush...>
        >>> g_names = set(b.name for b in py27_g)
        >>> u'brushlib-test/basic' in g_names
        True
        >>> u'brushlib-test/fancy_\U0001f308\U0001f984\u2728' in g_names
        True

        """

        with zipfile.ZipFile(path) as zf:

            # In Py2, when the entry was saved without the 0x800 flag
            # namelist() will return it as bytes, not unicode.  We only
            # want Unicode strings.
            names = []
            for name in zf.namelist():
                if isinstance(name, bytes):
                    name = name.decode("utf-8")
                names.append(name)

            readme = None
            if _BRUSHPACK_README in names:
                readme = zf.read(_BRUSHPACK_README).decode("utf-8")

            if _BRUSHPACK_ORDERCONF not in names:
                raise InvalidBrushpack(C_(
                    "brushpack import failure messages",
                    u"No file named “{order_conf_file}”. "
                    u"This is not a brushpack."
                ).format(
                    order_conf_file = _BRUSHPACK_ORDERCONF,
                ))
            groups = _parse_order_conf(zf.read(_BRUSHPACK_ORDERCONF))

            new_brushes = []
            for brushes in groups.values():
                for brush in brushes:
                    if brush not in new_brushes:
                        new_brushes.append(brush)
            logger.info(
                "%d different brushes found in %r of brushpack",
                len(new_brushes),
                _BRUSHPACK_ORDERCONF,
            )

            # Validate file content. The names in order.conf and the
            # brushes found in the zip must match. This should catch
            # encoding screwups, everything should be a unicode object.
            for brush in new_brushes:
                if brush + '.myb' not in names:
                    raise InvalidBrushpack(C_(
                        "brushpack import failure messages",
                        u"Brush “{brush_name}” is "
                        u"listed in “{order_conf_file}”, "
                        u"but it does not exist in the zipfile."
                    ).format(
                        brush_name = brush,
                        order_conf_file = _BRUSHPACK_ORDERCONF,
                    ))
            for name in names:
                if name.endswith('.myb'):
                    brush = name[:-4]
                    if brush not in new_brushes:
                        raise InvalidBrushpack(C_(
                            "brushpack import failure messages",
                            u"Brush “{brush_name}” exists in the zipfile, "
                            u"but it is not listed in “{order_conf_file}”."
                        ).format(
                            brush_name = brush,
                            order_conf_file = _BRUSHPACK_ORDERCONF,
                        ))
            if readme and window:
                answer = dialogs.confirm_brushpack_import(
                    basename(path), window, readme,
                )
                if answer == Gtk.ResponseType.REJECT:
                    return set()

            do_overwrite = False
            do_ask = True
            renamed_brushes = {}
            imported_groups = set()
            for groupname, brushes in groups.items():
                managed_brushes = self.get_group_brushes(groupname)
                if managed_brushes:
                    answer = dialogs.DONT_OVERWRITE_THIS
                    if window:
                        answer = dialogs.confirm_rewrite_group(
                            window, translate_group_name(groupname),
                            translate_group_name(DELETED_BRUSH_GROUP),
                        )
                    if answer == dialogs.CANCEL:
                        return set()
                    elif answer == dialogs.OVERWRITE_THIS:
                        self.delete_group(groupname)
                    elif answer == dialogs.DONT_OVERWRITE_THIS:
                        i = 0
                        old_groupname = groupname
                        while groupname in self.groups:
                            i += 1
                            groupname = old_groupname + '#%d' % i
                    managed_brushes = self.get_group_brushes(groupname)
                imported_groups.add(groupname)

                for brushname in brushes:
                    # extract the brush from the zip
                    assert (brushname + '.myb') in zf.namelist()
                    # Support for utf-8 ZIP filenames that don't have
                    # the utf-8 bit set.
                    brushname_utf8 = utf8(brushname)
                    try:
                        myb_data = zf.read(brushname + u'.myb')
                    except KeyError:
                        myb_data = zf.read(brushname_utf8 + b'.myb')
                    try:
                        preview_data = zf.read(brushname + u'_prev.png')
                    except KeyError:
                        preview_data = zf.read(brushname_utf8 + b'_prev.png')
                    # in case we have imported that brush already in a
                    # previous group, but decided to rename it
                    if brushname in renamed_brushes:
                        brushname = renamed_brushes[brushname]
                    # possibly ask how to import the brush file
                    # (if we didn't already)
                    b = self.get_brush_by_name(brushname)
                    if brushname in new_brushes:
                        new_brushes.remove(brushname)
                        if b:
                            existing_preview_pixbuf = b.preview
                            if do_ask and window:
                                answer = dialogs.confirm_rewrite_brush(
                                    window, brushname, existing_preview_pixbuf,
                                    preview_data,
                                )
                                if answer == dialogs.CANCEL:
                                    break
                                elif answer == dialogs.OVERWRITE_ALL:
                                    do_overwrite = True
                                    do_ask = False
                                elif answer == dialogs.OVERWRITE_THIS:
                                    do_overwrite = True
                                    do_ask = True
                                elif answer == dialogs.DONT_OVERWRITE_THIS:
                                    do_overwrite = False
                                    do_ask = True
                                elif answer == dialogs.DONT_OVERWRITE_ANYTHING:
                                    do_overwrite = False
                                    do_ask = False
                            # find a new name (if requested)
                            brushname_old = brushname
                            i = 0
                            while not do_overwrite and b:
                                i += 1
                                brushname = brushname_old + u'#%d' % i
                                renamed_brushes[brushname_old] = brushname
                                b = self.get_brush_by_name(brushname)

                        if not b:
                            b = ManagedBrush(self, brushname)

                        # write to disk and reload brush (if overwritten)
                        prefix = b._get_fileprefix(saving=True)
                        with open(prefix + '.myb', 'wb') as myb_f:
                            myb_f.write(myb_data)
                        with open(prefix + '_prev.png', 'wb') as preview_f:
                            preview_f.write(preview_data)
                        b.load()
                    # finally, add it to the group
                    if b not in managed_brushes:
                        managed_brushes.append(b)
                    self.brushes_changed(managed_brushes)

        if DELETED_BRUSH_GROUP in self.groups:
            # remove deleted brushes that are in some group again
            self.delete_group(DELETED_BRUSH_GROUP)
        return imported_groups

    def export_group(self, group, filename):
        """Exports a group to a brushpack zipfile.

        :param unicode|str group: Name of the group to save.
        :param unicode|str filename: Path to a .zip file to create.

        >>> with BrushManager._mock() as (bm, tmpdir):
        ...     group = list(bm.groups)[0]
        ...     zipname = os.path.join(tmpdir, group + u"zip")
        ...     bm.export_group(group, zipname)
        ...     with zipfile.ZipFile(zipname, mode="r") as zf:
        ...         assert len(zf.namelist()) > 0
        ...         assert u"order.conf" in zf.namelist()

        """
        brushes = self.get_group_brushes(group)
        order_conf = b'Group: %s\n' % utf8(group)
        with zipfile.ZipFile(filename, mode='w') as zf:
            for brush in brushes:
                prefix = brush._get_fileprefix()
                zf.write(prefix + u'.myb', brush.name + u'.myb')
                zf.write(prefix + u'_prev.png', brush.name + u'_prev.png')
                order_conf += utf8(brush.name + "\n")
            zf.writestr(u'order.conf', order_conf)

    ## Brush lookup / access

    def get_brush_by_name(self, name):
        """Gets a ManagedBrush by its name.

        Slow method, should not be called too often.

        >>> with BrushManager._mock() as (bm, tmpdir):
        ...     brush1 = bm.get_brush_by_name(u"classic/pen")
        >>> brush1 # doctest: +ELLIPSIS
        <ManagedBrush...>

        """
        # FIXME: speed up, use a dict.
        for group, brushes in self.groups.items():
            for b in brushes:
                if b.name == name:
                    return b

    def is_in_brushlist(self, brush):
        """Returns whether this brush is in some brush group's list."""
        for group, brushes in self.groups.items():
            if brush in brushes:
                return True
        return False

    def get_parent_brush(self, brush=None, brushinfo=None):
        """Gets the parent `ManagedBrush` for a brush or a `BrushInfo`.
        """
        if brush is not None:
            brushinfo = brush.brushinfo
        if brushinfo is None:
            raise RuntimeError("One of `brush` or `brushinfo` must be defined")
        parent_name = brushinfo.get_string_property("parent_brush_name")
        if parent_name is None:
            return None
        else:
            parent_brush = self.get_brush_by_name(parent_name)
            if parent_brush is None:
                return None
            return parent_brush

    ## Brush order within groups, order.conf

    def _brushes_modified_cb(self, bm, brushes):
        """Saves the brush order when it changes."""
        self.save_brushorder()

    def save_brushorder(self):
        """Save the user's chosen brush order to disk.

        >>> with BrushManager._mock() as (bm, tmpdir):
        ...     bm.save_brushorder()

        """

        path = os.path.join(self.user_brushpath, u'order.conf')
        with open(path, 'wb') as f:
            f.write(utf8(u'# this file saves brush groups and order\n'))
            for group, brushes in self.groups.items():
                f.write(utf8(u'Group: {}\n'.format(group)))
                for b in brushes:
                    f.write(utf8(b.name + u'\n'))

    ## The selected brush

    def select_brush(self, brush):
        """Selects a ManagedBrush, highlights it, & updates the live brush.

        :param brush: brush to select
        :type brush: BrushInfo

        """
        if brush is None:
            brush = self.get_default_brush()

        brushinfo = brush.brushinfo
        if not self.is_in_brushlist(brush):
            # select parent brush instead, but keep brushinfo
            parent = self.get_parent_brush(brush=brush)
            if parent is not None:
                brush = parent

        self.selected_brush = brush
        if self.app is not None:
            self.app.preferences['brushmanager.selected_brush'] = brush.name

        # Notify subscribers. Takes care of updating the live
        # brush, amongst other things
        self.brush_selected(brush, brushinfo)

    def clone_selected_brush(self, name):
        """Clones the current and selected brush into a new `BrushInfo`.

        Creates a new ManagedBrush based on the selected brush in the brushlist
        and the currently active lib.brush. The brush settings are copied from
        the active brush, and the preview is copied from the currently selected
        BrushInfo.

        """
        if self.app is None:
            raise ValueError("No app. BrushManager in test mode?")
        clone = ManagedBrush(self, name, persistent=False)
        clone.brushinfo = self.app.brush.clone()
        clone.preview = self.selected_brush.preview
        parent = self.selected_brush.name
        clone.brushinfo.set_string_property("parent_brush_name", parent)
        return clone

    def _brush_selected_cb(self, bm, brush, brushinfo):
        """Internal callback: User just picked a brush preset.

        Called when the user changes to a brush preset somehow (e.g.
        from a shortcut or the brush panel). Makes sure a
        brush-dependant tool (e.g. Freehand, Connected Lines, etc.) is
        selected.

        """
        if not self.app:
            return
        self.app.doc.modes.pop_to_behaviour(gui.mode.Behavior.PAINT_BRUSH)

    ## Device-specific brushes

    def store_brush_for_device(self, device_name, managed_brush):
        """Records a brush as associated with an input device.

        :param device_name: name of an input device
        :type device_name: str
        :param managed_brush: the brush to associate
        :type managed_brush: ManagedBrush

        Normally the brush will be cloned first, since it will be given a new
        name. However, if the brush has a 'name' attribute of None, it will
        *not* be cloned and just modified in place and stored.

        """
        brush = managed_brush
        if brush.name is not None:
            brush = brush.clone()
        brush.name = unicode(
            _DEVBRUSH_NAME_PREFIX + _device_name_uuid(device_name))
        self._brush_by_device[device_name] = brush

    def fetch_brush_for_device(self, device_name):
        """Fetches the brush associated with an input device."""
        if not device_name:
            return None

        if device_name not in self._brush_by_device:
            self._brush_by_device[device_name] = None

            names = (
                _device_name_uuid(device_name),
                _quote_device_name(device_name),  # for backward compatibility
            )
            for name in names:
                path = os.path.join(
                    self.user_brushpath, _DEVBRUSH_NAME_PREFIX + name + '.myb')
                if not os.path.isfile(path):
                    continue

                try:
                    b = ManagedBrush(
                        self, unicode(_DEVBRUSH_NAME_PREFIX + name),
                        persistent=True)
                except IOError as e:
                    logger.warn("%r: %r (ignored)", name, e)
                else:
                    self._brush_by_device[device_name] = b

                break

        assert device_name in self._brush_by_device
        return self._brush_by_device[device_name]

    def save_brushes_for_devices(self):
        """Saves the device/brush associations to disk."""
        for devbrush in self._brush_by_device.values():
            if devbrush is not None:
                devbrush.save()

    ## Brush history

    def _input_stroke_ended_cb(self, doc, event):
        """Update brush usage history at the end of an input stroke."""
        if self.app is None:
            raise ValueError("No app. BrushManager in test mode?")
        wb_info = self.app.brush
        wb_parent_name = wb_info.settings.get("parent_brush_name")
        # Remove the to-be-added brush from the history if it's already in it
        if wb_parent_name:
            # Favour "same parent" as the main measure of identity,
            # when it's defined.
            for i, hb in enumerate(self.history):
                hb_info = hb.brushinfo
                hb_parent_name = hb_info.settings.get("parent_brush_name")
                if wb_parent_name == hb_parent_name:
                    del self.history[i]
                    break
        else:
            # Otherwise, fall back to matching on the brush dynamics.
            # Many old .ORA files have pickable strokes in their layer map
            # which don't nominate a parent.
            for i, hb in enumerate(self.history):
                hb_info = hb.brushinfo
                if wb_info.matches(hb_info):
                    del self.history[i]
                    break
        # Append the working brush to the history, and trim it to length
        nb = ManagedBrush(self, name=None, persistent=False)
        nb.brushinfo = wb_info.clone()
        nb.preview = self.selected_brush.preview
        self.history.append(nb)
        while len(self.history) > _BRUSH_HISTORY_SIZE:
            del self.history[0]
        # Rename the history brushes so they save to the right files.
        for i, hb in enumerate(self.history):
            hb.name = u"%s%d" % (_BRUSH_HISTORY_NAME_PREFIX, i)

    def save_brush_history(self):
        """Saves the brush usage history to disk."""
        for brush in self.history:
            brush.save()

    ## Brush groups

    def get_group_brushes(self, group):
        """Get a group's active brush list.

        If the group does not exist, it will be created.

        :param str group: Name of the group to fetch
        :returns: The active list of `ManagedBrush`es.
        :rtype: list

        The returned list is owned by the BrushManager. You can modify
        it, but you'll have to do your own notifications.

        See also: groups_changed(), brushes_changed().

        """
        if group not in self.groups:
            brushes = []
            self.groups[group] = brushes
            self.groups_changed()
            self.save_brushorder()
        return self.groups[group]

    def create_group(self, new_group):
        """Creates a new brush group

        :param group: Name of the group to create
        :type group: str
        :rtype: empty list, owned by the BrushManager

        Returns the newly created group as a(n empty) list.

        """
        return self.get_group_brushes(new_group)

    def rename_group(self, old_group, new_group):
        """Renames a group.

        :param old_group: Name of the group to assign the new name to.
        :type old_group: str
        :param new_group: New name for the group.
        :type new_group: str

        """
        brushes = self.create_group(new_group)
        brushes += self.groups[old_group]
        self.delete_group(old_group)

    def delete_group(self, group):
        """Deletes a group.

        :param group: Name of the group to delete
        :type group: str

        Orphaned brushes will be placed into `DELETED_BRUSH_GROUP`, which
        will be created if necessary.

        """

        homeless_brushes = self.groups[group]
        del self.groups[group]

        for brushes in self.groups.values():
            for b2 in brushes:
                if b2 in homeless_brushes:
                    homeless_brushes.remove(b2)

        if homeless_brushes:
            deleted_brushes = self.get_group_brushes(DELETED_BRUSH_GROUP)
            for b in homeless_brushes:
                deleted_brushes.insert(0, b)
            self.brushes_changed(deleted_brushes)
        self.brushes_changed(homeless_brushes)
        self.groups_changed()
        self.save_brushorder()


class ManagedBrush(object):
    """User-facing representation of a brush's settings.

    Managed brushes have a name, a preview image, and brush settings.
    The settings and the preview are loaded on demand.
    They cannot be selected or painted with directly,
    but their settings can be loaded into the running app:
    see `Brushmanager.select_brush()`.

    """

    def __init__(self, brushmanager, name=None, persistent=False):
        """Construct, with a ref back to its BrushManager.

        Normally clients won't construct ManagedBrushes directly.
        Instead, use the groups dict in the BrushManager for access to
        all the brushes loaded from the user and stock brush folders.

        >>> with BrushManager._mock() as (bm, tmpdir):
        ...     for gname, gbrushes in bm.groups.items():
        ...         for b in gbrushes:
        ...             assert isinstance(b, ManagedBrush)
        ...             b.load()

        """

        super(ManagedBrush, self).__init__()
        self.bm = brushmanager
        self._preview = None
        self._brushinfo = BrushInfo(default_overrides={
            'paint_mode': self.bm.default_pigment_setting
        })

        #: The brush's relative filename, sans extension.
        self.name = name

        #: If True, this brush is stored in the filesystem.
        self.persistent = persistent

        # If True, this brush is fully initialized, ready to paint with.
        self._settings_loaded = False

        # Change detection for on-disk files.
        self._settings_mtime = None
        self._preview_mtime = None

        # Files are loaded later,
        # but throw an exception now if they don't exist.
        if persistent:
            self._get_fileprefix()
            assert self.name is not None

    def loaded(self):
        return self._settings_loaded

    ## Preview image: loaded on demand

    def get_preview(self):
        """Gets a preview image for the brush

        For persistent brushes, this loads the disk preview; otherwise a
        fairly slow automated brush preview is used.

        >>> with BrushManager._mock() as (bm, tmpdir):
        ...     b = ManagedBrush(bm, name=None, persistent=False)
        ...     b.get_preview()   # doctest: +ELLIPSIS
        <GdkPixbuf.Pixbuf...>

        The results are cached in RAM.

        >>> with BrushManager._mock() as (bm, tmpdir):
        ...     imported = bm.import_brushpack(_TEST_BRUSHPACK_PY27)
        ...     assert(imported)
        ...     pixbufs1 = []
        ...     for gn in sorted(bm.groups.keys()):
        ...         gbs = bm.groups[gn]
        ...         for b in gbs:
        ...             pixbufs1.append(b)
        ...     pixbufs2 = []
        ...     for gn in sorted(bm.groups.keys()):
        ...         gbs = bm.groups[gn]
        ...         for b in gbs:
        ...             pixbufs2.append(b)
        >>> len(pixbufs1) == len(pixbufs2)
        True
        >>> all([p1 is p2 for (p1, p2) in zip(pixbufs1, pixbufs2)])
        True
        >>> pixbufs1 == pixbufs2
        True

        """
        if self._preview is None and self.name:
            self._load_preview()
        if self._preview is None:
            brushinfo = self.get_brushinfo()
            self._preview = drawutils.render_brush_preview_pixbuf(brushinfo)
        return self._preview

    def set_preview(self, pixbuf):
        self._preview = pixbuf

    preview = property(get_preview, set_preview)

    ## Text fields

    @property
    def description(self):
        """Short, user-facing tooltip description for the brush.

        >>> with BrushManager._mock() as (bm, tmpdir):
        ...     for gn, gbs in bm.groups.items():
        ...         for b in gbs:
        ...             assert isinstance(b.description, unicode)
        ...             b.description = u"junk"
        ...             assert isinstance(b.description, unicode)
        ...             b.save()

        """
        return self.brushinfo.get_string_property("description")

    @description.setter
    def description(self, s):
        self.brushinfo.set_string_property("description", s)

    @property
    def notes(self):
        """Longer, brush developer's notes field for a brush.

        >>> with BrushManager._mock() as (bm, tmpdir):
        ...     imp = bm.import_brushpack(_TEST_BRUSHPACK_PY27)
        ...     imp_g = list(imp)[0]
        ...     for b in bm.groups[imp_g]:
        ...         assert isinstance(b.notes, unicode)
        ...         b.notes = u"junk note"
        ...         assert isinstance(b.notes, unicode)
        ...         b.save()

        """
        return self.brushinfo.get_string_property("notes")

    @notes.setter
    def notes(self, s):
        self.brushinfo.set_string_property("notes", s)

    ## Brush settings: loaded on demand

    def get_brushinfo(self):
        self._ensure_settings_loaded()
        return self._brushinfo

    def set_brushinfo(self, brushinfo):
        self._brushinfo = brushinfo

    brushinfo = property(get_brushinfo, set_brushinfo)

    ## Display

    def __repr__(self):
        if self._brushinfo.settings:
            pname = self._brushinfo.get_string_property("parent_brush_name")
            return "<ManagedBrush %r p=%s>" % (self.name, pname)
        else:
            return "<ManagedBrush %r (settings not loaded yet)>" % self.name

    def get_display_name(self):
        """Gets a displayable name for the brush."""
        if self.bm.is_in_brushlist(self):  # FIXME: get rid of this check
            dname = self.name
        else:
            dname = self.brushinfo.get_string_property("parent_brush_name")
        if dname is None:
            return _("Unknown Brush")
        return dname.replace("_", " ")

    ## Cloning

    def clone(self, name):
        """Clone this brush, and give it a new name.

        Creates a new brush with all the settings of this brush,
        assigning it a new name

        """
        clone = ManagedBrush(self.bm)
        self.clone_into(clone, name=name)
        return clone

    def clone_into(self, target, name):
        "Copies all brush settings into another brush, giving it a new name"
        self._ensure_settings_loaded()
        target.brushinfo = self.brushinfo.clone()
        if self.bm.is_in_brushlist(self):  # FIXME: get rid of this check!
            target.brushinfo.set_string_property(
                "parent_brush_name", self.name,
            )
        target.preview = self.preview
        target.name = name

    ## File save/load helpers

    def _get_fileprefix(self, saving=False):
        """Returns the filesystem prefix to use when saving or loading.

        :param saving: caller wants a prefix to save to
        :type saving: bool
        :rtype: unicode

        This assigns ``self.name`` if it isn't defined.

        Files are stored with the returned prefix,
        with the extension ".myb" for brush data
        and "_prev.myb" for preview images.

        If `saving` is true, intermediate directories will be created,
        and the returned prefix will always contain the user brushpath.
        Otherwise the prefix you get depends on
        whether a stock brush exists and
        whether a user brush with the same name does not.

        See also `delete_from_disk()`.

        """
        prefix = u'b'
        user_bp = os.path.realpath(self.bm.user_brushpath)
        stock_bp = os.path.realpath(self.bm.stock_brushpath)
        if user_bp == stock_bp:
            # working directly on brush collection, use different prefix
            prefix = u's'

        # Construct a new, unique name if the brush is not yet named
        if not self.name:
            i = 0
            while True:
                self.name = u'%s%03d' % (prefix, i)
                a = os.path.join(self.bm.user_brushpath, self.name + u'.myb')
                b = os.path.join(self.bm.stock_brushpath, self.name + u'.myb')
                if not os.path.isfile(a) and not os.path.isfile(b):
                    break
                i += 1
        assert isinstance(self.name, unicode)

        # Always save to the user brush path.
        prefix = os.path.join(self.bm.user_brushpath, self.name)
        if saving:
            if u'/' in self.name:
                d = os.path.dirname(prefix)
                if not os.path.isdir(d):
                    os.makedirs(d)
            return prefix

        # Loading: try user first, then stock
        if not os.path.isfile(prefix + u'.myb'):
            prefix = os.path.join(self.bm.stock_brushpath, self.name)
        if not os.path.isfile(prefix + u'.myb'):
            raise IOError('brush "%s" not found' % self.name)
        return prefix

    def _remember_mtimes(self):
        prefix = self._get_fileprefix()
        try:
            preview_file = prefix + '_prev.png'
            self._preview_mtime = os.path.getmtime(preview_file)
        except OSError:
            logger.exception("Failed to update preview file access time")
            self._preview_mtime = None
        try:
            settings_file = prefix + '.myb'
            self._settings_mtime = os.path.getmtime(settings_file)
        except OSError:
            logger.exception("Failed to update settings file access time")
            self._settings_mtime = None

    ## Saving and deleting

    def save(self):
        """Saves the brush's settings and its preview"""
        prefix = self._get_fileprefix(saving=True)
        # Save preview
        if self.preview.get_has_alpha():
            # Remove alpha
            # Previous mypaint versions would display an empty image
            w, h = PREVIEW_W, PREVIEW_H
            tmp = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, False,
                                       8, w, h)
            tmp.fill(0xffffffff)
            self.preview.composite(tmp, 0, 0, w, h, 0, 0, 1, 1,
                                   GdkPixbuf.InterpType.BILINEAR, 255)
            self.preview = tmp
        preview_filename = prefix + '_prev.png'
        logger.debug("Saving brush preview to %r", preview_filename)
        lib.pixbuf.save(self.preview, preview_filename, "png")
        # Save brush settings
        brushinfo = self.brushinfo.clone()
        settings_filename = prefix + '.myb'
        logger.debug("Saving brush settings to %r", settings_filename)
        with open(settings_filename, 'w') as settings_fp:
            settings_fp.write(brushinfo.save_to_string())
        # Record metadata
        self._remember_mtimes()

    def delete_from_disk(self):
        """Tries to delete the files for this brush from disk.

        :rtype: bool

        Returns True if the disk files can no longer be loaded. Stock brushes
        cannot be deleted, but if a user brush is hiding a stock brush with the
        same name, then although this method will remove the files describing
        the user brush, the stock brush is left intact. In this case, False is
        returned (because a load() attempt will now load the stock brush - and
        in fact has just done so).

        """

        prefix = os.path.join(self.bm.user_brushpath, self.name)
        if os.path.isfile(prefix + '.myb'):
            os.remove(prefix + '_prev.png')
            os.remove(prefix + '.myb')
            try:
                self.load()
            except IOError:
                # Files are no longer there, and no stock files with the
                # same name could be loaded.
                return True
            else:
                # User brush was hiding a stock brush with the same name.
                return False
        # Stock brushes cannot be deleted.
        return False

    ## Loading and reloading

    def load(self):
        """Loads the brush's preview and settings from disk."""
        if self.name is None:
            warn("Attempt to load an unnamed brush, don't do that.",
                 RuntimeWarning, 2)
            return
        self._load_preview()
        self._load_settings()

    def _load_preview(self):
        """Loads the brush preview as pixbuf into the brush."""
        assert self.name
        prefix = self._get_fileprefix()
        filename = prefix + '_prev.png'
        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
        except Exception:
            logger.exception("Failed to load preview pixbuf, will fall back "
                             "to default")
            pixbuf = None
        self._preview = pixbuf
        self._remember_mtimes()

    def _load_settings(self):
        """Loads the brush settings/dynamics from disk."""
        prefix = self._get_fileprefix()
        filename = prefix + '.myb'
        with open(filename) as fp:
            brushinfo_str = fp.read()
        try:
            self._brushinfo.load_from_string(brushinfo_str)
        except Exception as e:
            logger.warning('Failed to load brush %r: %s', filename, e)
            self._brushinfo.load_defaults()
        self._remember_mtimes()
        self._settings_loaded = True
        if self.bm.is_in_brushlist(self):  # FIXME: get rid of this check
            self._brushinfo.set_string_property("parent_brush_name", None)
        self.persistent = True

    def _has_changed_on_disk(self):
        prefix = self._get_fileprefix()
        if self._preview_mtime != os.path.getmtime(prefix + '_prev.png'):
            return True
        if self._settings_mtime != os.path.getmtime(prefix + '.myb'):
            return True
        return False

    def reload_if_changed(self):
        if self._settings_mtime is None:
            return
        if self._preview_mtime is None:
            return
        if not self.name:
            return
        if not self._has_changed_on_disk():
            return False
        logger.info('Brush %r has changed on disk, reloading it.',
                    self.name)
        self.load()
        return True

    def _ensure_settings_loaded(self):
        """Ensures the brush's settings are loaded, if persistent"""
        if self.persistent and not self._settings_loaded:
            logger.debug("Loading %r...", self)
            self.load()
            assert self._settings_loaded


class InvalidBrushpack (Exception):
    """Raised when brushpacks cannot be imported."""


## Module testing

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2012 by Richard Jones
# Copyright (C) 2012-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# NOTE: Much of this was written before the new modes system.
# NOTE: The InteractionMode stuff was sort of bolted on after the fact.

# TODO: Understand it more; devolve specifics to specific subclasses.
# TODO: Document stuff properly.


## Imports

from __future__ import division, print_function
import math
import logging
from gettext import gettext as _

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GLib

from .curve import CurveWidget
import gui.mode
import gui.cursor
from lib.pycompat import xrange


logger = logging.getLogger(__name__)

## Module constants

# internal-name, display-name, constant, minimum, default, maximum, tooltip
_LINE_MODE_SETTINGS_LIST = [
    ['entry_pressure', _('Entrance Pressure'), False, 0.0001, 0.3, 1.0,
     _("Stroke entrance pressure for line tools")],
    ['midpoint_pressure', _('Midpoint Pressure'), False, 0.0001, 1.0, 1.0,
     _("Mid-Stroke pressure for line tools")],
    ['exit_pressure', _('Exit Pressure'), False, 0.0001, 0.3, 1.0,
     _("Stroke exit pressure for line tools")],
    ['line_head', _('Head'), False, 0.0001, 0.25, 1.0,
     _("Stroke lead-in end")],
    ['line_tail', _('Tail'), False, 0.0001, 0.75, 1.0,
     _("Stroke trail-off beginning")],
]


## Line pressure settings

class LineModeSettings (object):
    """Manage GtkAdjustments for tweaking LineMode settings.

    An instance resides in the main application singleton. Changes to the
    adjustments are reflected into the app preferences.
    """

    def __init__(self, app):
        """Initializer; initial settings are loaded from the app prefs"""
        object.__init__(self)
        self.app = app
        self.adjustments = {}  #: Dictionary of GtkAdjustments
        self.observers = []  #: List of callbacks
        self._idle_srcid = None
        self._changed_settings = set()
        for line_list in _LINE_MODE_SETTINGS_LIST:
            cname, name, const, min_, default, max_, tooltip = line_list
            prefs_key = "linemode.%s" % cname
            value = float(self.app.preferences.get(prefs_key, default))
            adj = Gtk.Adjustment(value=value, lower=min_, upper=max_,
                                 step_increment=0.01, page_increment=0.1)
            adj.connect("value-changed", self._value_changed_cb, prefs_key)
            self.adjustments[cname] = adj

    def _value_changed_cb(self, adj, prefs_key):
        # Direct GtkAdjustment callback for a single adjustment being changed.
        value = float(adj.get_value())
        self.app.preferences[prefs_key] = value
        self._changed_settings.add(prefs_key)
        if self._idle_srcid is None:
            self._idle_srcid = GLib.idle_add(self._values_changed_idle_cb)

    def _values_changed_idle_cb(self):
        # Aggregate, idle-state callback for multiple adjustments being changed
        # in a single event. Queues redraws, and runs observers. The curve sets
        # multiple settings at once, and we might as well not queue too many
        # redraws.
        if self._idle_srcid is not None:
            current_mode = self.app.doc.modes.top
            if isinstance(current_mode, LineModeBase):
                # Redraw last_line when settings are adjusted
                # in the adjustment Curve
                GLib.idle_add(current_mode.redraw_line_cb)
            for func in self.observers:
                func(self._changed_settings)
            self._changed_settings = set()
            self._idle_srcid = None
        return False


class LineModeCurveWidget (CurveWidget):
    """Graph of pressure by distance, tied to the central LineModeSettings"""

    _SETTINGS_COORDINATE = [('entry_pressure', (0, 1)),
                            ('midpoint_pressure', (1, 1)),
                            ('exit_pressure', (3, 1)),
                            ('line_head', (1, 0)),
                            ('line_tail', (2, 0))]

    def __init__(self):
        from gui.application import get_app
        self.app = get_app()
        CurveWidget.__init__(self, npoints=4, ylockgroups=((1, 2),),
                             changed_cb=self._changed_cb)
        self.app.line_mode_settings.observers.append(self._adjs_changed_cb)
        self._update()

    def _adjs_changed_cb(self, changed):
        logger.debug("Updating curve (changed: %r)", changed)
        self._update()

    def _update(self, from_defaults=False):
        if from_defaults:
            self.points = [(0.0, 0.2), (0.33, 0.5), (0.66, 0.5), (1.0, 0.33)]
        for setting, coord_pair in self._SETTINGS_COORDINATE:
            if not from_defaults:
                adj = self.app.line_mode_settings.adjustments[setting]
                value = adj.get_value()
            else:
                # TODO: move this case into the base settings object
                defaults = [a[4] for a in _LINE_MODE_SETTINGS_LIST
                            if a[0] == setting]
                assert len(defaults) == 1
                value = defaults[0]
            index, subindex = coord_pair
            if not setting.startswith('line'):
                value = 1.0 - value
            if subindex == 0:
                coord = (value, self.points[index][1])
            else:
                coord = (self.points[index][0], value)
            self.set_point(index, coord)
        if from_defaults:
            self._changed_cb(self)
        self.queue_draw()

    def _changed_cb(self, curve):
        """Updates the linemode pressure settings when the curve is altered"""
        for setting, coord_pair in self._SETTINGS_COORDINATE:
            index, subindex = coord_pair
            value = self.points[index][subindex]
            if not setting.startswith('line'):
                value = 1.0 - value
            value = max(0.0001, value)
            adj = self.app.line_mode_settings.adjustments[setting]
            adj.set_value(value)


## Options UI

class LineModeOptionsWidget (gui.mode.PaintingModeOptionsWidgetBase):
    """Options widget for geometric line modes"""

    def init_specialized_widgets(self, row=0):
        curve = LineModeCurveWidget()
        curve.set_size_request(175, 125)
        self._curve = curve
        exp = Gtk.Expander()
        exp.set_label(_(u"Pressure variation…"))
        exp.set_use_markup(False)
        exp.add(curve)
        self.attach(exp, 0, row, 2, 1)
        row += 1
        return row

    def reset_button_clicked_cb(self, button):
        super(LineModeOptionsWidget, self).reset_button_clicked_cb(button)
        self._curve._update(from_defaults=True)


## Interaction modes for making lines

class LineModeBase (gui.mode.ScrollableModeMixin,
                    gui.mode.BrushworkModeMixin,
                    gui.mode.DragMode):
    """Draws geometric lines (base class)"""

    ## Class constants

    _OPTIONS_WIDGET = None

    ## Class configuration.

    permitted_switch_actions = set([
        "PanViewMode",
        "ZoomViewMode",
        "RotateViewMode",
    ])

    pointer_behavior = gui.mode.Behavior.PAINT_CONSTRAINED
    scroll_behavior = gui.mode.Behavior.CHANGE_VIEW

    @property
    def active_cursor(self):
        cursor_name = gui.cursor.Name.PENCIL
        if not self._line_possible:
            cursor_name = gui.cursor.Name.FORBIDDEN_EVERYWHERE
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME,
            cursor_name
        )

    @classmethod
    def get_name(cls):
        return _(u"Lines and Curves")

    def get_usage(self):
        # TRANSLATORS: users should never see this message
        return _(u"Generic line/curve mode")

    @property
    def inactive_cursor(self):
        cursor_name = gui.cursor.Name.CROSSHAIR_OPEN_PRECISE
        if not self._line_possible:
            cursor_name = gui.cursor.Name.FORBIDDEN_EVERYWHERE
        return self.doc.app.cursors.get_action_cursor(
            self.ACTION_NAME,
            cursor_name
        )

    unmodified_persist = True
    permitted_switch_actions = set(
        ['RotateViewMode', 'ZoomViewMode', 'PanViewMode']
        + gui.mode.BUTTON_BINDING_ACTIONS)

    # FIXME: all of the logic resides in the base class, for historical
    # reasons, and is decided by line_mode. The differences should be
    # factored out to the user-facing mode subclasses at some point.
    line_mode = None

    ## Initialization

    def __init__(self, **kwds):
        """Initialize"""
        super(LineModeBase, self).__init__(**kwds)
        self.app = None
        self.last_line_data = None
        self.idle_srcid = None
        self._line_possible = False

    ## InteractionMode/DragMode implementation

    def enter(self, doc, **kwds):
        """Enter the mode.

        If modifiers are held when the mode is entered, the mode is a oneshot
        mode and is popped from the mode stack automatically at the end of the
        drag. Without modifiers, line modes may be continued, and some
        subclasses offer additional options for adjusting control points.

        """
        super(LineModeBase, self).enter(doc, **kwds)
        self.app = self.doc.app
        rootstack = self.doc.model.layer_stack
        rootstack.current_path_updated += self._update_cursors
        rootstack.layer_properties_changed += self._update_cursors
        self._update_cursors()

    def leave(self, **kwds):
        rootstack = self.doc.model.layer_stack
        rootstack.current_path_updated -= self._update_cursors
        rootstack.layer_properties_changed -= self._update_cursors
        return super(LineModeBase, self).leave(**kwds)

    def _update_cursors(self, *_ignored):
        if self.in_drag:
            return   # defer update to the end of the drag
        layer = self.doc.model.layer_stack.current
        self._line_possible = layer.get_paintable()
        self.doc.tdw.set_override_cursor(self.inactive_cursor)

    def drag_start_cb(self, tdw, event):
        super(LineModeBase, self).drag_start_cb(tdw, event)
        if self._line_possible:
            self.start_command(self.initial_modifiers)

    def drag_update_cb(self, tdw, event, dx, dy):
        if self._line_possible:
            self.update_position(event.x, event.y)
            if self.idle_srcid is None:
                self.idle_srcid = GLib.idle_add(self._drag_idle_cb)
        return super(LineModeBase, self).drag_update_cb(tdw, event, dx, dy)

    def drag_stop_cb(self, tdw):
        if self._line_possible:
            self.idle_srcid = None
            self.stop_command()
        self._update_cursors()
        return super(LineModeBase, self).drag_stop_cb(tdw)

    def _drag_idle_cb(self):
        # Updates the on-screen line during drags.
        if self.idle_srcid is not None:
            self.idle_srcid = None
            self.process_line()

    def checkpoint(self, flush=True, **kwargs):
        # Only push outstanding changes to the document's undo stack on
        # a request for a flushing sync. Without this, users can't curve
        # the previously drawn line. See also inkmode.py and
        # https://github.com/mypaint/mypaint/issues/262
        if flush:
            super(LineModeBase, self).checkpoint(flush=True, **kwargs)

    ## Options panel

    def get_options_widget(self):
        """Get the (base class singleton) options widget"""
        cls = LineModeBase
        if cls._OPTIONS_WIDGET is None:
            widget = LineModeOptionsWidget()
            cls._OPTIONS_WIDGET = widget
        return cls._OPTIONS_WIDGET

    ### Draw dynamic Line, Curve, or Ellipse

    def start_command(self, modifier):
        # :param modifier: the keyboard modifiers which ere in place
        #                   when the mode was created

        active_tdw = self.app.doc.tdw.__class__.get_active_tdw()
        assert active_tdw is not None
        if active_tdw is self.app.scratchpad_doc.tdw:
            self.model = self.app.scratchpad_doc.model
            self.tdw = self.app.scratchpad_doc.tdw
        else:
            # unconditionally
            self.model = self.app.doc.model
            self.tdw = self.app.doc.tdw

        self.done = False
        self.brushwork_begin(self.model, abrupt=False,
                             description=self.get_name())
        layer = self.model.layer_stack.current

        x, y, kbmods = self.local_mouse_state()
        # ignore the modifier used to start this action (don't make it
        # change the action)
        self.invert_kbmods = modifier
        kbmods ^= self.invert_kbmods  # invert using bitwise xor
        shift = kbmods & Gdk.ModifierType.SHIFT_MASK

        # line_mode is the type of line to be drawn eg. "EllipseMode"
        self.mode = self.line_mode
        assert self.mode is not None

        # Ignore slow_tracking.
        # There are some other sttings that interfere
        # with the workings of the Line Tools,
        # but slowtracking is the main one.
        self.adj = self.app.brush_adjustment['slow_tracking']
        self.slow_tracking = self.adj.get_value()
        self.adj.set_value(0)

        # Throughout this module these conventions are used:
        # sx, sy = starting point
        # ex, ey = end point
        # kx, ky = curve point from last line
        # lx, ly = last point from InteractionMode update
        self.sx, self.sy = x, y
        self.lx, self.ly = x, y

        if self.mode == "EllipseMode":
            # Rotation angle of ellipse.
            self.angle = 90
            # Vector to measure any rotation from.
            # Assigned when ratation begins.
            self.ellipse_vec = None
            return
        # If not Ellipse, command must be Straight Line or Sequence
        # First check if the user intends to Curve an existing Line
        if shift:
            last_line = self.last_line_data
            last_stroke = layer.get_last_stroke_info()
            if last_line is not None:
                if last_line[1] == last_stroke:
                    self.mode = last_line[0]
                    self.sx, self.sy = last_line[2], last_line[3]
                    self.ex, self.ey = last_line[4], last_line[5]
                    if self.mode == "CurveLine2":
                        length_a = distance(x, y, self.sx, self.sy)
                        length_b = distance(x, y, self.ex, self.ey)
                        self.flip = length_a > length_b
                        if self.flip:
                            self.kx, self.ky = last_line[6], last_line[7]
                        else:
                            self.kx, self.ky = last_line[8], last_line[9]
                    self.model.undo()
                    self.process_line()
                    return

        if self.mode == "SequenceMode":
            if not self.tdw.last_painting_pos:
                return
            else:
                self.sx, self.sy = self.tdw.last_painting_pos

    def update_position(self, x, y):
        self.lx, self.ly = self.tdw.display_to_model(x, y)

    def stop_command(self):
        # End mode
        self.done = True
        x, y = self.process_line()
        self.brushwork_commit(self.model, abrupt=False)
        cmd = self.mode
        self.record_last_stroke(cmd, x, y)

    def record_last_stroke(self, cmd, x, y):
        last_line = None
        self.tdw.last_painting_pos = x, y
        # FIXME: should probably not set that from here

        layer = self.model.layer_stack.current
        last_stroke = layer.get_last_stroke_info()
        sx, sy = self.sx, self.sy

        if cmd == "CurveLine1":
            last_line = [
                "CurveLine2", last_stroke,
                sx, sy,
                self.ex, self.ey,
                x, y, x, y,
            ]
            self.tdw.last_painting_pos = self.ex, self.ey

        if cmd == "CurveLine2":
            if self.flip:
                last_line = [
                    cmd, last_stroke,
                    sx, sy,
                    self.ex, self.ey,
                    self.kx, self.ky, x, y,
                ]
            else:
                last_line = [
                    cmd, last_stroke,
                    sx, sy,
                    self.ex, self.ey,
                    x, y, self.kx, self.ky,
                ]
            self.tdw.last_painting_pos = self.ex, self.ey

        if cmd == "StraightMode" or cmd == "SequenceMode":
            last_line = ["CurveLine1", last_stroke, sx, sy, x, y]

        if cmd == "EllipseMode":
            last_line = [cmd, last_stroke, sx, sy, x, y, self.angle]
            self.tdw.last_painting_pos = sx, sy

        self.last_line_data = last_line
        self.adj.set_value(self.slow_tracking)
        self.model.brush.reset()

    def local_mouse_state(self, last_update=False):
        tdw_win = self.tdw.get_window()
        display = self.tdw.get_display()
        devmgr = display and display.get_device_manager() or None
        coredev = devmgr and devmgr.get_client_pointer() or None
        if coredev and tdw_win:
            win_, x, y, kbmods = tdw_win.get_device_position_double(coredev)
        else:
            x, y, kbmods = (0., 0., Gdk.ModifierType(0))
        if last_update:
            return self.lx, self.ly, kbmods
        x, y = self.tdw.display_to_model(x, y)
        return x, y, kbmods

    def process_line(self):
        sx, sy = self.sx, self.sy
        x, y, kbmods = self.local_mouse_state(last_update=True)
        kbmods ^= self.invert_kbmods  # invert using bitwise xor
        ctrl = kbmods & Gdk.ModifierType.CONTROL_MASK
        shift = kbmods & Gdk.ModifierType.SHIFT_MASK

        if self.mode == "CurveLine1":
            self.dynamic_curve_1(x, y, sx, sy, self.ex, self.ey)

        elif self.mode == "CurveLine2":
            ex, ey = self.ex, self.ey
            kx, ky = self.kx, self.ky
            if not self.flip:
                self.dynamic_curve_2(x, y, sx, sy, ex, ey, kx, ky)
            else:
                self.dynamic_curve_2(kx, ky, sx, sy, ex, ey, x, y)

        elif self.mode == "EllipseMode":
            constrain = False
            if ctrl:
                x, y = constrain_to_angle(x, y, sx, sy)
                constrain = True
            if shift:
                self.ellipse_rotation_angle(x, y, sx, sy, constrain)
            else:
                self.ellipse_vec = None
            self.dynamic_ellipse(x, y, sx, sy)

        else:  # if "StraightMode" or "SequenceMode"
            if ctrl or shift:
                x, y = constrain_to_angle(x, y, sx, sy)
            self.dynamic_straight_line(x, y, sx, sy)
        return x, y

    def ellipse_rotation_angle(self, x, y, sx, sy, constrain):
        x1, y1 = normal(sx, sy, x, y)
        if self.ellipse_vec is None:
            self.ellipse_vec = x1, y1
            self.last_angle = self.angle
        x2, y2 = self.ellipse_vec
        px, py = perpendicular(x2, y2)
        pangle = get_angle(x1, y1, px, py)
        angle = get_angle(x1, y1, x2, y2)
        if pangle > 90.0:
            angle = 360 - angle
        angle += self.last_angle
        if constrain:
            angle = constraint_angle(angle)
        self.angle = angle

    ### Line Functions

    # Straight Line
    def dynamic_straight_line(self, x, y, sx, sy):
        self.brush_prep(sx, sy)
        entry_p, midpoint_p, junk, prange2, head, tail = self.line_settings()
        # Beginning
        length, nx, ny = length_and_normal(sx, sy, x, y)
        mx, my = multiply_add(sx, sy, nx, ny, 0.25)
        self._stroke_to(mx, my, entry_p)
        # Middle start
        # length = length/2
        mx, my = multiply_add(sx, sy, nx, ny, head * length)
        self._stroke_to(mx, my, midpoint_p)
        # Middle end
        mx, my = multiply_add(sx, sy, nx, ny, tail * length)
        self._stroke_to(mx, my, midpoint_p)
        # End
        self._stroke_to(x, y, self.exit_pressure)

    # Ellipse
    def dynamic_ellipse(self, x, y, sx, sy):
        points_in_curve = 360
        x1, y1 = difference(sx, sy, x, y)
        x1, y1, sin, cos = starting_point_for_ellipse(x1, y1, self.angle)
        rx, ry = point_in_ellipse(x1, y1, sin, cos, 0)
        self.brush_prep(sx+rx, sy+ry)
        entry_p, midpoint_p, prange1, prange2, h, t = self.line_settings()
        head = points_in_curve * h
        head_range = int(head)+1
        tail = points_in_curve * t
        tail_range = int(tail)+1
        tail_length = points_in_curve - tail
        # Beginning
        px, py = point_in_ellipse(x1, y1, sin, cos, 1)
        length, nx, ny = length_and_normal(rx, ry, px, py)
        mx, my = multiply_add(rx, ry, nx, ny, 0.25)
        self._stroke_to(sx+mx, sy+my, entry_p)
        pressure = abs(1/head * prange1 + entry_p)
        self._stroke_to(sx+px, sy+py, pressure)
        for degree in xrange(2, head_range):
            px, py = point_in_ellipse(x1, y1, sin, cos, degree)
            pressure = abs(degree/head * prange1 + entry_p)
            self._stroke_to(sx+px, sy+py, pressure)
        # Middle
        for degree in xrange(head_range, tail_range):
            px, py = point_in_ellipse(x1, y1, sin, cos, degree)
            self._stroke_to(sx+px, sy+py, midpoint_p)
        # End
        for degree in xrange(tail_range, points_in_curve+1):
            px, py = point_in_ellipse(x1, y1, sin, cos, degree)
            pressure = abs((degree-tail)/tail_length * prange2 + midpoint_p)
            self._stroke_to(sx+px, sy+py, pressure)

    def dynamic_curve_1(self, cx, cy, sx, sy, ex, ey):
        self.brush_prep(sx, sy)
        self.draw_curve_1(cx, cy, sx, sy, ex, ey)

    def dynamic_curve_2(self, cx, cy, sx, sy, ex, ey, kx, ky):
        self.brush_prep(sx, sy)
        self.draw_curve_2(cx, cy, sx, sy, ex, ey, kx, ky)

    # Curve Straight Line
    # Found this page helpful:
    # http://www.caffeineowl.com/graphics/2d/vectorial/bezierintro.html
    def draw_curve_1(self, cx, cy, sx, sy, ex, ey):
        points_in_curve = 100
        entry_p, midpoint_p, prange1, prange2, h, t = self.line_settings()
        mx, my = midpoint(sx, sy, ex, ey)
        length, nx, ny = length_and_normal(mx, my, cx, cy)
        cx, cy = multiply_add(mx, my, nx, ny, length*2)
        x1, y1 = difference(sx, sy, cx, cy)
        x2, y2 = difference(cx, cy, ex, ey)
        head = points_in_curve * h
        head_range = int(head)+1
        tail = points_in_curve * t
        tail_range = int(tail)+1
        tail_length = points_in_curve - tail
        # Beginning
        px, py = point_on_curve_1(1, cx, cy, sx, sy, x1, y1, x2, y2)
        length, nx, ny = length_and_normal(sx, sy, px, py)
        bx, by = multiply_add(sx, sy, nx, ny, 0.25)
        self._stroke_to(bx, by, entry_p)
        pressure = abs(1/head * prange1 + entry_p)
        self._stroke_to(px, py, pressure)
        for i in xrange(2, head_range):
            px, py = point_on_curve_1(i, cx, cy, sx, sy, x1, y1, x2, y2)
            pressure = abs(i/head * prange1 + entry_p)
            self._stroke_to(px, py, pressure)
        # Middle
        for i in xrange(head_range, tail_range):
            px, py = point_on_curve_1(i, cx, cy, sx, sy, x1, y1, x2, y2)
            self._stroke_to(px, py, midpoint_p)
        # End
        for i in xrange(tail_range, points_in_curve+1):
            px, py = point_on_curve_1(i, cx, cy, sx, sy, x1, y1, x2, y2)
            pressure = abs((i-tail)/tail_length * prange2 + midpoint_p)
            self._stroke_to(px, py, pressure)

    def draw_curve_2(self, cx, cy, sx, sy, ex, ey, kx, ky):
        points_in_curve = 100
        self.brush_prep(sx, sy)
        entry_p, midpoint_p, prange1, prange2, h, t = self.line_settings()
        mx, my = (cx+sx+ex+kx)/4.0, (cy+sy+ey+ky)/4.0
        length, nx, ny = length_and_normal(mx, my, cx, cy)
        cx, cy = multiply_add(mx, my, nx, ny, length*2)
        length, nx, ny = length_and_normal(mx, my, kx, ky)
        kx, ky = multiply_add(mx, my, nx, ny, length*2)
        x1, y1 = difference(sx, sy, cx, cy)
        x2, y2 = difference(cx, cy, kx, ky)
        x3, y3 = difference(kx, ky, ex, ey)
        head = points_in_curve * h
        head_range = int(head)+1
        tail = points_in_curve * t
        tail_range = int(tail)+1
        tail_length = points_in_curve - tail
        # Beginning
        px, py = point_on_curve_2(1, cx, cy, sx, sy, kx, ky,
                                  x1, y1, x2, y2, x3, y3)
        length, nx, ny = length_and_normal(sx, sy, px, py)
        bx, by = multiply_add(sx, sy, nx, ny, 0.25)
        self._stroke_to(bx, by, entry_p)
        pressure = abs(1/head * prange1 + entry_p)
        self._stroke_to(px, py, pressure)
        for i in xrange(2, head_range):
            px, py = point_on_curve_2(i, cx, cy, sx, sy, kx, ky,
                                      x1, y1, x2, y2, x3, y3)
            pressure = abs(i/head * prange1 + entry_p)
            self._stroke_to(px, py, pressure)
        # Middle
        for i in xrange(head_range, tail_range):
            px, py = point_on_curve_2(i, cx, cy, sx, sy, kx, ky,
                                      x1, y1, x2, y2, x3, y3)
            self._stroke_to(px, py, midpoint_p)
        # End
        for i in xrange(tail_range, points_in_curve+1):
            px, py = point_on_curve_2(i, cx, cy, sx, sy, kx, ky,
                                      x1, y1, x2, y2, x3, y3)
            pressure = abs((i-tail)/tail_length * prange2 + midpoint_p)
            self._stroke_to(px, py, pressure)

    def _stroke_to(self, x, y, pressure):
        # FIXME add control for time, similar to inktool
        duration = 1.0

        self.stroke_to(self.model, duration, x, y, pressure, 0.0, 0.0,
                       self.doc.tdw.scale, self.doc.tdw.rotation, 0.0,
                       auto_split=False)

    def brush_prep(self, sx, sy):
        # Send brush to where the stroke will begin
        self.model.brush.reset()
        self.brushwork_rollback(self.model)

        self.stroke_to(self.model, 10.0, sx, sy, 0.0, 0.0, 0.0,
                       self.doc.tdw.scale, self.doc.tdw.rotation, 0.0,
                       auto_split=False)

    ## Line mode settings

    @property
    def entry_pressure(self):
        adj = self.app.line_mode_settings.adjustments["entry_pressure"]
        return adj.get_value()

    @property
    def midpoint_pressure(self):
        adj = self.app.line_mode_settings.adjustments["midpoint_pressure"]
        return adj.get_value()

    @property
    def exit_pressure(self):
        adj = self.app.line_mode_settings.adjustments["exit_pressure"]
        return adj.get_value()

    @property
    def head(self):
        adj = self.app.line_mode_settings.adjustments["line_head"]
        return adj.get_value()

    @property
    def tail(self):
        adj = self.app.line_mode_settings.adjustments["line_tail"]
        return adj.get_value()

    def line_settings(self):
        p1 = self.entry_pressure
        p2 = self.midpoint_pressure
        p3 = self.exit_pressure
        if self.head == 0.0001:
            p1 = p2
        prange1 = p2 - p1
        prange2 = p3 - p2
        return p1, p2, prange1, prange2, self.head, self.tail

    def redraw_line_cb(self):
        # Redraws the line when the line_mode_settings change
        last_line = self.last_line_data
        if last_line is not None:
            current_layer = self.model.layer_stack.current
            last_stroke = current_layer.get_last_stroke_info()
            if last_line[1] is last_stroke:
                # ignore slow_tracking
                self.done = True
                self.adj = self.app.brush_adjustment['slow_tracking']
                self.slow_tracking = self.adj.get_value()
                self.adj.set_value(0)
                self.brushwork_rollback(self.model)
                self.model.undo()
                command = last_line[0]
                self.sx, self.sy = last_line[2], last_line[3]
                self.ex, self.ey = last_line[4], last_line[5]
                x, y = self.ex, self.ey
                if command == "EllipseMode":
                    self.angle = last_line[6]
                    self.dynamic_ellipse(self.ex, self.ey,
                                         self.sx, self.sy)
                if command == "CurveLine1":
                    self.dynamic_straight_line(self.ex, self.ey,
                                               self.sx, self.sy)
                    command = "StraightMode"
                if command == "CurveLine2":
                    x, y = last_line[6], last_line[7]
                    self.kx, self.ky = last_line[8], last_line[9]
                    if (x, y) == (self.kx, self.ky):
                        self.dynamic_curve_1(x, y, self.sx, self.sy,
                                             self.ex, self.ey)
                        command = "CurveLine1"
                    else:
                        self.flip = False
                        self.dynamic_curve_2(x, y, self.sx, self.sy,
                                             self.ex, self.ey,
                                             self.kx, self.ky)
                self.model.sync_pending_changes()
                self.record_last_stroke(command, x, y)


class StraightMode (LineModeBase):
    ACTION_NAME = "StraightMode"
    line_mode = "StraightMode"

    @classmethod
    def get_name(cls):
        return _(u"Lines and Curves")

    def get_usage(self):
        return _(u"Draw straight lines; Shift adds curves, "
                 "Ctrl constrains angle")


class SequenceMode (LineModeBase):
    ACTION_NAME = "SequenceMode"
    line_mode = "SequenceMode"

    @classmethod
    def get_name(cls):
        return _(u"Connected Lines")

    def get_usage(cls):
        return _("Draw a sequence of lines; Shift adds curves, "
                 "Ctrl constrains angle")


class EllipseMode (LineModeBase):
    ACTION_NAME = "EllipseMode"
    line_mode = "EllipseMode"

    @classmethod
    def get_name(cls):
        return _(u"Ellipses and Circles")

    def get_usage(self):
        return _(u"Draw ellipses; Shift rotates, Ctrl constrains ratio/angle")


## Curve Math
def point_on_curve_1(t, cx, cy, sx, sy, x1, y1, x2, y2):
    ratio = t/100.0
    x3, y3 = multiply_add(sx, sy, x1, y1, ratio)
    x4, y4 = multiply_add(cx, cy, x2, y2, ratio)
    x5, y5 = difference(x3, y3, x4, y4)
    x, y = multiply_add(x3, y3, x5, y5, ratio)
    return x, y


def point_on_curve_2(t, cx, cy, sx, sy, kx, ky, x1, y1, x2, y2, x3, y3):
    ratio = t/100.0
    x4, y4 = multiply_add(sx, sy, x1, y1, ratio)
    x5, y5 = multiply_add(cx, cy, x2, y2, ratio)
    x6, y6 = multiply_add(kx, ky, x3, y3, ratio)
    x1, y1 = difference(x4, y4, x5, y5)
    x2, y2 = difference(x5, y5, x6, y6)
    x4, y4 = multiply_add(x4, y4, x1, y1, ratio)
    x5, y5 = multiply_add(x5, y5, x2, y2, ratio)
    x1, y1 = difference(x4, y4, x5, y5)
    x, y = multiply_add(x4, y4, x1, y1, ratio)
    return x, y


## Ellipse Math
def starting_point_for_ellipse(x, y, rotate):
    # Rotate starting point
    r = math.radians(rotate)
    sin = math.sin(r)
    cos = math.cos(r)
    x, y = rotate_ellipse(x, y, cos, sin)
    return x, y, sin, cos


def point_in_ellipse(x, y, r_sin, r_cos, degree):
    # Find point in ellipse
    r2 = math.radians(degree)
    cos = math.cos(r2)
    sin = math.sin(r2)
    x = x * cos
    y = y * sin
    # Rotate Ellipse
    x, y = rotate_ellipse(y, x, r_sin, r_cos)
    return x, y


def rotate_ellipse(x, y, sin, cos):
    x1, y1 = multiply(x, y, sin)
    x2, y2 = multiply(x, y, cos)
    x = x2 - y1
    y = y2 + x1
    return x, y


## Vector Math
def get_angle(x1, y1, x2, y2):
    dot = dot_product(x1, y1, x2, y2)
    if abs(dot) < 1.0:
        angle = math.acos(dot) * 180/math.pi
    else:
        angle = 0.0
    return angle


def constrain_to_angle(x, y, sx, sy):
    length, nx, ny = length_and_normal(sx, sy, x, y)
    # dot = nx*1 + ny*0 therefore nx
    angle = math.acos(nx) * 180/math.pi
    angle = constraint_angle(angle)
    ax, ay = angle_normal(ny, angle)
    x = sx + ax*length
    y = sy + ay*length
    return x, y


def constraint_angle(angle):
    n = angle//15
    n1 = n*15
    rem = angle - n1
    if rem < 7.5:
        angle = n*15.0
    else:
        angle = (n+1)*15.0
    return angle


def angle_normal(ny, angle):
    if ny < 0.0:
        angle = 360.0 - angle
    radians = math.radians(angle)
    x = math.cos(radians)
    y = math.sin(radians)
    return x, y


def length_and_normal(x1, y1, x2, y2):
    x, y = difference(x1, y1, x2, y2)
    length = vector_length(x, y)
    if length == 0.0:
        x, y = 0.0, 0.0
    else:
        x, y = x/length, y/length
    return length, x, y


def normal(x1, y1, x2, y2):
    junk, x, y = length_and_normal(x1, y1, x2, y2)
    return x, y


def vector_length(x, y):
    length = math.sqrt(x*x + y*y)
    return length


def distance(x1, y1, x2, y2):
    x, y = difference(x1, y1, x2, y2)
    length = vector_length(x, y)
    return length


def dot_product(x1, y1, x2, y2):
    return x1*x2 + y1*y2


def multiply_add(x1, y1, x2, y2, d):
    x3, y3 = multiply(x2, y2, d)
    x, y = add(x1, y1, x3, y3)
    return x, y


def multiply(x, y, d):
    # Multiply vector
    x = x*d
    y = y*d
    return x, y


def add(x1, y1, x2, y2):
    # Add vectors
    x = x1+x2
    y = y1+y2
    return x, y


def difference(x1, y1, x2, y2):
    # Difference in x and y between two points
    x = x2-x1
    y = y2-y1
    return x, y


def midpoint(x1, y1, x2, y2):
    # Midpoint between to points
    x = (x1+x2)/2.0
    y = (y1+y2)/2.0
    return x, y


def perpendicular(x1, y1):
    # Swap x and y, then flip one sign to give vector at 90 degree
    x = -y1
    y = x1
    return x, y



# This file is part of MyPaint.
# Copyright (C) 2011-2018 by the MyPaint Development Team.
# Copyright (C) 2008-2012 by Martin Renold <martinxyz@gmx.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


## Imports

from __future__ import division, print_function
import contextlib
from logging import getLogger

from lib.gibindings import GdkPixbuf
from lib.gibindings import Gdk
import cairo

from . import mypaintlib
from . import helpers
from lib.eotf import eotf
import lib.surface
from lib.surface import TileAccessible, TileBlittable
from lib.errors import AllocationError
from lib.gettext import C_
import lib.feedback

logger = getLogger(__name__)


## Module consts

TILE_SIZE = N = mypaintlib.TILE_SIZE

_POSSIBLE_OOM_USERTEXT = C_(
    "user-facing error texts",
    u"Unable to construct a vital internal object. "
    u"Your system may not have enough memory to perform "
    u"this operation."
)


## Class defs

class Surface (TileAccessible, TileBlittable):
    """Wrapper for a GdkPixbuf, with memory accessible by tile.

    Wraps a GdkPixbuf.Pixbuf (8 bit RGBU or RGBA data) with memory also
    accessible per-tile, compatible with tiledsurface.Surface.

    This class converts between linear 8bit RGB(A) and tiled RGBA storage. It
    is used for rendering updates, but also for save/load.

    """

    def __init__(self, x, y, w, h, data=None):
        super(Surface, self).__init__()
        assert w > 0 and h > 0

        # We create and use a pixbuf enlarged to the tile boundaries
        # internally.  Variables ex, ey, ew, eh and epixbuf store the
        # enlarged version.

        self.x, self.y, self.w, self.h = x, y, w, h
        tx = self.tx = x // N
        ty = self.ty = y // N
        self.ex = tx * N
        self.ey = ty * N
        tw = (x + w - 1) // N - tx + 1
        th = (y + h - 1) // N - ty + 1

        self.ew = tw * N
        self.eh = th * N

        assert self.ew >= w and self.eh >= h
        assert self.ex <= x and self.ey <= y

        # Tile-aligned pixbuf: also accessible by tile
        try:
            self.epixbuf = GdkPixbuf.Pixbuf.new(
                GdkPixbuf.Colorspace.RGB, True, 8,
                self.ew, self.eh
            )
        except Exception as te:
            logger.exception("GdkPixbuf.Pixbuf.new() failed")
            raise AllocationError(_POSSIBLE_OOM_USERTEXT)
        if self.epixbuf is None:
            logger.error("GdkPixbuf.Pixbuf.new() returned NULL")
            raise AllocationError(_POSSIBLE_OOM_USERTEXT)

        # External subpixbuf, also accessible by tile.
        dx = x - self.ex
        dy = y - self.ey
        try:
            self.pixbuf = self.epixbuf.new_subpixbuf(dx, dy, w, h)
        except Exception as te:
            logger.exception("GdkPixbuf.Pixbuf.new_subpixbuf() failed")
            raise AllocationError(_POSSIBLE_OOM_USERTEXT)
        if self.pixbuf is None:
            logger.error("GdkPixbuf.Pixbuf.new_subpixbuf() returned NULL")
            raise AllocationError(_POSSIBLE_OOM_USERTEXT)

        assert self.ew <= w + (2 * N) - 2
        assert self.eh <= h + (2 * N) - 2

        self.epixbuf.fill(0x00000000)  # keep undefined regions transparent

        # Make it accessible by tile
        arr = helpers.gdkpixbuf2numpy(self.epixbuf)
        assert len(arr) > 0

        discard_transparent = False

        if data is not None:
            dst = arr[dy:dy + h, dx:dx + w, :]
            if data.shape[2] == 4:
                dst[:, :, :] = data
                discard_transparent = True
            else:
                assert data.shape[2] == 3
                # no alpha channel
                dst[:, :, :3] = data
                dst[:, :, 3] = 255

        # Build (tx,ty)-indexed access struct
        self.tile_memory_dict = {}
        for ty in range(th):
            for tx in range(tw):
                buf = arr[ty * N:(ty + 1) * N, tx * N:(tx + 1) * N, :]
                if discard_transparent and not buf[:, :, 3].any():
                    continue
                self.tile_memory_dict[(self.tx + tx, self.ty + ty)] = buf

    def get_bbox(self):
        return lib.surface.get_tiles_bbox(self.get_tiles())

    def get_tiles(self):
        return self.tile_memory_dict

    @contextlib.contextmanager
    def tile_request(self, tx, ty, readonly):
        """Access memory by tile (lib.surface.TileAccessible impl.)"""
        numpy_tile = self._get_tile_numpy(tx, ty, readonly)
        yield numpy_tile
        self._set_tile_numpy(tx, ty, numpy_tile, readonly)

    def _get_tile_numpy(self, tx, ty, readonly):
        return self.tile_memory_dict[(tx, ty)]

    def _set_tile_numpy(self, tx, ty, arr, readonly):
        pass  # Data can be modified directly, no action needed

    def blit_tile_into(self, dst, dst_has_alpha, tx, ty):
        # (used mainly for loading transparent PNGs)
        assert dst_has_alpha is True
        assert dst.dtype == 'uint16', '16 bit dst expected'
        src = self.tile_memory_dict[(tx, ty)]
        assert src.shape[2] == 4, 'alpha required'
        mypaintlib.tile_convert_rgba8_to_rgba16(src, dst, eotf())

    @contextlib.contextmanager
    def cairo_request(self):
        """Access via a temporary Cairo context.

        Modifications are copied back into the backing pixbuf when the
        context manager finishes.

        """
        # Make a Cairo surface copy of the subpixbuf
        surf = cairo.ImageSurface(
            cairo.FORMAT_ARGB32,
            self.pixbuf.get_width(), self.pixbuf.get_height(),
        )
        cr = cairo.Context(surf)
        Gdk.cairo_set_source_pixbuf(cr, self.pixbuf, 0, 0)
        cr.paint()

        # User can modify its content with Cairo operations
        yield cr

        # Put the modified data back into the tile-aligned pixbuf.
        dx = self.x - self.ex
        dy = self.y - self.ey
        pixbuf = Gdk.pixbuf_get_from_surface(surf, 0, 0, self.w, self.h)
        pixbuf.copy_area(0, 0, self.w, self.h, self.epixbuf, dx, dy)


def render_as_pixbuf(surface, x=None, y=None, w=None, h=None,
                     alpha=False, mipmap_level=0,
                     progress=None,
                     **kwargs):
    """Renders a surface within a given rectangle as a GdkPixbuf

    :param lib.surface.TileBlittable surface: source surface
    :param int x:
    :patrm int y:
    :param int w:
    :param int h: coords of the rendering rectangle. Must all be set.
    :param bool alpha:
    :param int mipmap_level:
    :param progress: Unsized UI progress feedback obj.
    :type progress: lib.feedback.Progress or None
    :param **kwargs: Keyword args are passed to ``surface.blit_tile_into()``
    :rtype: GdkPixbuf
    :raises: lib.errors.AllocationError
    :raises: MemoryError

    """
    if None in (x, y, w, h):
        x, y, w, h = surface.get_bbox()
    s = Surface(x, y, w, h)
    s_tiles = list(s.get_tiles())
    if not progress:
        progress = lib.feedback.Progress()
    progress.items = len(s_tiles)
    tn = 0
    for tx, ty in s_tiles:
        with s.tile_request(tx, ty, readonly=False) as dst:
            surface.blit_tile_into(dst, alpha, tx, ty,
                                   mipmap_level=mipmap_level,
                                   **kwargs)
            if tn % lib.surface.TILES_PER_CALLBACK == 0:
                progress.completed(tn)
            tn += 1
    progress.close()
    return s.pixbuf



# This file is part of MyPaint.
# Copyright (C) 2013-2018 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Palette: user-defined lists of color swatches"""

# TODO: Make palettes part of the model, save as part of ORA documents.


## Imports

from __future__ import division, print_function

import re
from copy import copy
import logging

from lib.helpers import clamp
from lib.observable import event
from lib.color import RGBColor
from lib.color import YCbCrColor
from lib.color import UIColor  # noqa
from lib.pycompat import unicode
from lib.pycompat import xrange
from lib.pycompat import PY3
from io import open

logger = logging.getLogger(__name__)

## Class and function defs


class Palette (object):
    """A flat list of color swatches, compatible with the GIMP

    As a (sideways-compatible) extension to the GIMP's format, MyPaint supports
    empty slots in the palette. These slots are represented by pure black
    swatches with the name ``__NONE__``.

    Palette objects expose the position within the palette of a current color
    match, which can be declared to be approximate or exact. This is used for
    highlighting the user concept of the "current color" in the GUI.

    Palette objects can be serialized in the GIMP's file format (the regular
    `unicode()` function on a Palette will do this too), or converted to and
    from a simpler JSON-ready representation for storing in the MyPaint prefs.
    Support for loading and saving via modal dialogs is defined here too.

    """

    ## Class-level constants
    _EMPTY_SLOT_ITEM = RGBColor(-1, -1, -1)
    _EMPTY_SLOT_NAME = "__NONE__"

    ## Construction, loading and saving

    def __init__(self, filehandle=None, filename=None, colors=None):
        """Instantiate, from a file or a sequence of colors

        :param filehandle: Filehandle to load.
        :param filename: Name of a file to load.
        :param colors: Iterable sequence of colors (lib.color.UIColor).

        The constructor arguments are mutually exclusive.  With no args
        specified, you get an empty palette.

          >>> Palette()
          <Palette colors=0, columns=0, name=None>

        Palettes can be generated from interpolations, which is handy for
        testing, at least.

          >>> cols = RGBColor(1,1,0).interpolate(RGBColor(1,0,1), 10)
          >>> Palette(colors=cols)
          <Palette colors=10, columns=0, name=None>

        """
        super(Palette, self).__init__()

        #: Number of columns. 0 means "natural flow"
        self._columns = 0
        #: List of named colors
        self._colors = []
        #: Name of the palette as a Unicode string, or None
        self._name = None
        #: Current position in the palette. None=no match; integer=index.
        self._match_position = None
        #: True if the current match is approximate
        self._match_is_approx = False

        # Clear and initialize
        self.clear(silent=True)
        if colors:
            for col in colors:
                col = self._copy_color_in(col)
                self._colors.append(col)
        elif filehandle:
            self.load(filehandle, silent=True)
        elif filename:
            with open(filename, "r", encoding="utf-8", errors="replace") as fp:
                self.load(fp, silent=True)

    def clear(self, silent=False):
        """Resets the palette to its initial state.

          >>> grey16 = RGBColor(1,1,1).interpolate(RGBColor(0,0,0), 16)
          >>> p = Palette(colors=grey16)
          >>> p.name = "Greyscale"
          >>> p.columns = 3
          >>> p                               # doctest: +ELLIPSIS
          <Palette colors=16, columns=3, name=...'Greyscale'>
          >>> p.clear()
          >>> p
          <Palette colors=0, columns=0, name=None>

        Fires the `info_changed()`, `sequence_changed()`, and `match_changed()`
        events, unless the `silent` parameter tests true.
        """
        self._colors = []
        self._columns = 0
        self._name = None
        self._match_position = None
        self._match_is_approx = False
        if not silent:
            self.info_changed()
            self.sequence_changed()
            self.match_changed()

    def load(self, filehandle, silent=False):
        """Load contents from a file handle containing a GIMP palette.

        :param filehandle: File-like object (.readline, line iteration)
        :param bool silent: If true, don't emit any events.

        >>> pal = Palette()
        >>> with open("palettes/MyPaint_Default.gpl", "r") as fp:
        ...     pal.load(fp)
        >>> len(pal) > 1
        True

        If the file format is incorrect, a RuntimeError will be raised.

        """
        comment_line_re = re.compile(r'^#')
        field_line_re = re.compile(r'^(\w+)\s*:\s*(.*)$')
        color_line_re = re.compile(r'^(\d+)\s+(\d+)\s+(\d+)\s*(?:\b(.*))$')
        fp = filehandle
        self.clear(silent=True)   # method fires events itself
        line = fp.readline()
        if line.strip() != "GIMP Palette":
            raise RuntimeError("Not a valid GIMP Palette")
        header_done = False
        line_num = 0
        for line in fp:
            line = line.strip()
            line_num += 1
            if line == '':
                continue
            if comment_line_re.match(line):
                continue
            if not header_done:
                match = field_line_re.match(line)
                if match:
                    key, value = match.groups()
                    key = key.lower()
                    if key == 'name':
                        self._name = value.strip()
                    elif key == 'columns':
                        self._columns = int(value)
                    else:
                        logger.warning("Unknown 'key:value' pair %r", line)
                    continue
                else:
                    header_done = True
            match = color_line_re.match(line)
            if not match:
                logger.warning("Expected 'R G B [Name]', not %r", line)
                continue
            r, g, b, col_name = match.groups()
            col_name = col_name.strip()
            r = clamp(int(r), 0, 0xff) / 0xff
            g = clamp(int(g), 0, 0xff) / 0xff
            b = clamp(int(b), 0, 0xff) / 0xff
            if r == g == b == 0 and col_name == self._EMPTY_SLOT_NAME:
                self.append(None)
            else:
                col = RGBColor(r, g, b)
                col.__name = col_name
                self._colors.append(col)
        if not silent:
            self.info_changed()
            self.sequence_changed()
            self.match_changed()

    def save(self, filehandle):
        """Saves the palette to an open file handle.

        :param filehandle: File-like object (.write suffices)

        >>> from lib.pycompat import PY3
        >>> if PY3:
        ...     from io import StringIO
        ... else:
        ...     from cStringIO import StringIO
        >>> fp = StringIO()
        >>> cols = RGBColor(1,.7,0).interpolate(RGBColor(.1,.1,.5), 16)
        >>> pal = Palette(colors=cols)
        >>> pal.save(fp)
        >>> fp.getvalue() == unicode(pal)
        True

        The file handle is not flushed, and is left open after the
        write.

        >>> fp.flush()
        >>> fp.close()

        """
        filehandle.write(unicode(self))

    def update(self, other):
        """Updates all details of this palette from another palette.

        Fires the `info_changed()`, `sequence_changed()`, and `match_changed()`
        events.
        """
        self.clear(silent=True)
        for col in other._colors:
            col = self._copy_color_in(col)
            self._colors.append(col)
        self._name = other._name
        self._columns = other._columns
        self.info_changed()
        self.sequence_changed()
        self.match_changed()

    ## Palette size and metadata

    def get_columns(self):
        """Get the number of columns (0 means unspecified)."""
        return self._columns

    def set_columns(self, n):
        """Set the number of columns (0 means unspecified)."""
        self._columns = int(n)
        self.info_changed()

    def get_name(self):
        """Gets the palette's name."""
        return self._name

    def set_name(self, name):
        """Sets the palette's name."""
        if name is not None:
            name = unicode(name)
        self._name = name
        self.info_changed()

    def __bool__(self):
        """Palettes never test false, regardless of their length.

        >>> p = Palette()
        >>> bool(p)
        True

        """
        return True

    def __len__(self):
        """Palette length is the number of color slots within it."""
        return len(self._colors)

    ## PY2/PY3 compat

    __nonzero__ = __bool__

    ## Match position marker

    def get_match_position(self):
        """Return the position of the current match (int or None)"""
        return self._match_position

    def set_match_position(self, i):
        """Sets the position of the current match (int or None)

        Fires `match_changed()` if the value is changed."""
        if i is not None:
            i = int(i)
            if i < 0 or i >= len(self):
                i = None
        if i != self._match_position:
            self._match_position = i
            self.match_changed()

    def get_match_is_approx(self):
        """Returns whether the current match is approximate."""
        return self._match_is_approx

    def set_match_is_approx(self, approx):
        """Sets whether the current match is approximate

        Fires match_changed() if the boolean value changes."""
        approx = bool(approx)
        if approx != self._match_is_approx:
            self._match_is_approx = approx
            self.match_changed()

    def match_color(self, col, exact=False, order=None):
        """Moves the match position to the color closest to the argument.

        :param col: The color to match.
        :type col: lib.color.UIColor
        :param exact: Only consider exact matches, and not near-exact or
                approximate matches.
        :type exact: bool
        :param order: a search order to use. Default is outwards from the
                match position, or in order if the match is unset.
        :type order: sequence or iterator of integer color indices.
        :returns: Whether the match succeeded.
        :rtype: bool

        By default, the matching algorithm favours exact or near-exact matches
        which are close to the current position. If the current position is
        unset, this search starts at 0. If there are no exact or near-exact
        matches, a looser approximate match will be used, again favouring
        matches with nearby positions.

          >>> red2blue = RGBColor(1, 0, 0).interpolate(RGBColor(0, 1, 1), 5)
          >>> p = Palette(colors=red2blue)
          >>> p.match_color(RGBColor(0.45, 0.45, 0.45))
          True
          >>> p.match_position
          2
          >>> p.match_is_approx
          True
          >>> p[p.match_position]
          <RGBColor r=0.5000, g=0.5000, b=0.5000>
          >>> p.match_color(RGBColor(0.5, 0.5, 0.5))
          True
          >>> p.match_is_approx
          False
          >>> p.match_color(RGBColor(0.45, 0.45, 0.45), exact=True)
          False
          >>> p.match_color(RGBColor(0.5, 0.5, 0.5), exact=True)
          True

        Fires the ``match_changed()`` event when changes happen.
        """
        if order is not None:
            search_order = order
        elif self.match_position is not None:
            search_order = _outwards_from(len(self), self.match_position)
        else:
            search_order = xrange(len(self))
        bestmatch_i = None
        bestmatch_d = None
        is_approx = True
        for i in search_order:
            c = self._colors[i]
            if c is self._EMPTY_SLOT_ITEM:
                continue
            # Closest exact or near-exact match by index distance (according to
            # the search_order). Considering near-exact matches as equivalent
            # to exact matches improves the feel of PaletteNext and
            # PalettePrev.
            if exact:
                if c == col:
                    bestmatch_i = i
                    is_approx = False
                    break
            else:
                d = _color_distance(col, c)
                if c == col or d < 0.06:
                    bestmatch_i = i
                    is_approx = False
                    break
                if bestmatch_d is None or d < bestmatch_d:
                    bestmatch_i = i
                    bestmatch_d = d
            # Measuring over a blend into solid equiluminant 0-chroma
            # grey for the orange #DA5D2E with an opaque but feathered
            # brush made huge, and picking just inside the point where the
            # palette widget begins to call it approximate:
            #
            # 0.05 is a difference only discernible (to me) by tilting LCD
            # 0.066 to 0.075 appears slightly greyer for large areas
            # 0.1 and above is very clearly distinct

        # If there are no exact or near-exact matches, choose the most similar
        # color anywhere in the palette.
        if bestmatch_i is not None:
            self._match_position = bestmatch_i
            self._match_is_approx = is_approx
            self.match_changed()
            return True
        return False

    def move_match_position(self, direction, refcol):
        """Move the match position in steps, matching first if needed.

        :param direction: Direction for moving, positive or negative
        :type direction: int:, ``1`` or ``-1``
        :param refcol: Reference color, used for initial matching when needed.
        :type refcol: UIColor
        :returns: the color newly matched, if the match position has changed
        :rtype: UIColor|NoneType

        Invoking this method when there's no current match position will select
        the color that's closest to the reference color, just like
        `match_color()`

        >>> greys = RGBColor(1,1,1).interpolate(RGBColor(0,0,0), 16)
        >>> pal = Palette(colors=greys)
        >>> refcol = RGBColor(0.5, 0.55, 0.45)
        >>> pal.move_match_position(-1, refcol)
        >>> pal.match_position
        7
        >>> pal.match_is_approx
        True

        When the current match is defined, but only an approximate match, this
        method converts it to an exact match but does not change its position.

          >>> pal.move_match_position(-1, refcol) is None
          False
          >>> pal.match_position
          7
          >>> pal.match_is_approx
          False

        When the match is initially exact, its position is stepped in the
        direction indicated, either by +1 or -1. Blank palette entries are
        skipped.

          >>> pal.move_match_position(-1, refcol) is None
          False
          >>> pal.match_position
          6
          >>> pal.match_is_approx
          False

        Fires ``match_position_changed()`` and ``match_is_approx_changed()`` as
        appropriate.  The return value is the newly matched color whenever this
        method produces a new exact match.

        """
        # Normalize direction
        direction = int(direction)
        if direction < 0:
            direction = -1
        elif direction > 0:
            direction = 1
        else:
            return None
        # If nothing is selected, pick the closest match without changing
        # the managed color.
        old_pos = self._match_position
        if old_pos is None:
            self.match_color(refcol)
            return None
        # Otherwise, refine the match, or step it in the requested direction.
        new_pos = None
        if self._match_is_approx:
            # Make an existing approximate match concrete.
            new_pos = old_pos
        else:
            # Index reflects a close or identical match.
            # Seek in the requested direction, skipping empty entries.
            pos = old_pos
            assert direction != 0
            pos += direction
            while pos < len(self._colors) and pos >= 0:
                if self._colors[pos] is not self._EMPTY_SLOT_ITEM:
                    new_pos = pos
                    break
                pos += direction
        # Update the palette index and the managed color.
        result = None
        if new_pos is not None:
            col = self._colors[new_pos]
            if col is not self._EMPTY_SLOT_ITEM:
                result = self._copy_color_out(col)
            self.set_match_position(new_pos)
            self.set_match_is_approx(False)
        return result

    ## Property-style access for setters and getters

    columns = property(get_columns, set_columns)
    name = property(get_name, set_name)
    match_position = property(get_match_position, set_match_position)
    match_is_approx = property(get_match_is_approx, set_match_is_approx)

    ## Color access

    def _copy_color_out(self, col):
        if col is self._EMPTY_SLOT_ITEM:
            return None
        result = RGBColor(color=col)
        result.__name = col.__name
        return result

    def _copy_color_in(self, col, name=None):
        if col is self._EMPTY_SLOT_ITEM or col is None:
            result = self._EMPTY_SLOT_ITEM
        else:
            if name is None:
                try:
                    name = col.__name
                except AttributeError:
                    pass
            if name is not None:
                name = unicode(name)
            result = RGBColor(color=col)
            result.__name = name
        return result

    def append(self, col, name=None, unique=False, match=False):
        """Appends a color, optionally setting a name for it.

        :param col: The color to append.
        :param name: Name of the color to insert.
        :param unique: If true, don't append if the color already exists
                in the palette. Only exact matches count.
        :param match: If true, set the match position to the
                appropriate palette entry.
        """
        col = self._copy_color_in(col, name)
        if unique:
            # Find the final exact match, if one is present
            for i in xrange(len(self._colors)-1, -1, -1):
                if col == self._colors[i]:
                    if match:
                        self._match_position = i
                        self._match_is_approx = False
                        self.match_changed()
                    return
        # Append new color, and select it if requested
        end_i = len(self._colors)
        self._colors.append(col)
        if match:
            self._match_position = end_i
            self._match_is_approx = False
            self.match_changed()
        self.sequence_changed()

    def insert(self, i, col, name=None):
        """Inserts a color, setting an optional name for it.

        :param i: Target index. `None` indicates appending a color.
        :param col: Color to insert. `None` indicates an empty slot.
        :param name: Name of the color to insert.

          >>> grey16 = RGBColor(1, 1, 1).interpolate(RGBColor(0, 0, 0), 16)
          >>> p = Palette(colors=grey16)
          >>> p.insert(5, RGBColor(1, 0, 0), name="red")
          >>> p
          <Palette colors=17, columns=0, name=None>
          >>> p[5]
          <RGBColor r=1.0000, g=0.0000, b=0.0000>

        Fires the `sequence_changed()` event. If the match position changes as
        a result, `match_changed()` is fired too.

        """
        col = self._copy_color_in(col, name)
        if i is None:
            self._colors.append(col)
        else:
            self._colors.insert(i, col)
            if self.match_position is not None:
                if self.match_position >= i:
                    self.match_position += 1
        self.sequence_changed()

    def reposition(self, src_i, targ_i):
        """Moves a color, or copies it to empty slots, or moves it the end.

        :param src_i: Source color index.
        :param targ_i: Source color index, or None to indicate the end.

        This operation performs a copy if the target is an empty slot, and a
        remove followed by an insert if the target slot contains a color.

          >>> grey16 = RGBColor(1, 1, 1).interpolate(RGBColor(0, 0, 0), 16)
          >>> p = Palette(colors=grey16)
          >>> p[5] = None           # creates an empty slot
          >>> p.match_position = 8
          >>> p[5] == p[0]
          False
          >>> p.reposition(0, 5)
          >>> p[5] == p[0]
          True
          >>> p.match_position
          8
          >>> p[5] = RGBColor(1, 0, 0)
          >>> p.reposition(14, 5)
          >>> p.match_position     # continues pointing to the same color
          9
          >>> len(p)       # repositioning doesn't change the length
          16

        Fires the `color_changed()` event for copies to empty slots, or
        `sequence_changed()` for moves. If the match position changes as a
        result, `match_changed()` is fired too.

        """
        assert src_i is not None
        if src_i == targ_i:
            return
        try:
            col = self._colors[src_i]
            assert col is not None  # just in case we change the internal repr
        except IndexError:
            return

        # Special case: just copy if the target is an empty slot
        match_pos = self.match_position
        if targ_i is not None:
            targ = self._colors[targ_i]
            if targ is self._EMPTY_SLOT_ITEM:
                self._colors[targ_i] = self._copy_color_in(col)
                self.color_changed(targ_i)
                # Copying from the matched color moves the match position.
                # Copying to the match position clears the match.
                if match_pos == src_i:
                    self.match_position = targ_i
                elif match_pos == targ_i:
                    self.match_position = None
                return

        # Normal case. Remove...
        self._colors.pop(src_i)
        moving_match = False
        updated_match = False
        if match_pos is not None:
            # Moving rightwards. Adjust for the pop().
            if targ_i is not None and targ_i > src_i:
                targ_i -= 1
            # Similar logic for the match position, but allow it to follow
            # the move if it started at the src position.
            if match_pos == src_i:
                match_pos = None
                moving_match = True
                updated_match = True
            elif match_pos > src_i:
                match_pos -= 1
                updated_match = True
        # ... then append or insert.
        if targ_i is None:
            self._colors.append(col)
            if moving_match:
                match_pos = len(self._colors) - 1
                updated_match = True
        else:
            self._colors.insert(targ_i, col)
            if match_pos is not None:
                if moving_match:
                    match_pos = targ_i
                    updated_match = True
                elif match_pos >= targ_i:
                    match_pos += 1
                    updated_match = True
        # Announce changes
        self.sequence_changed()
        if updated_match:
            self.match_position = match_pos
            self.match_changed()

    def pop(self, i):
        """Removes a color, returning it.

        Fires the `match_changed()` event if the match index changes as a
        result of the removal, and `sequence_changed()` if a color was removed,
        prior to its return.
        """
        i = int(i)
        try:
            col = self._colors.pop(i)
        except IndexError:
            return
        if self.match_position == i:
            self.match_position = None
        elif self.match_position > i:
            self.match_position -= 1
        self.sequence_changed()
        return self._copy_color_out(col)

    def get_color(self, i):
        """Looks up a color by its list index."""
        if i is None:
            return None
        try:
            col = self._colors[i]
            return self._copy_color_out(col)
        except IndexError:
            return None

    def __getitem__(self, i):
        return self.get_color(i)

    def __setitem__(self, i, col):
        self._colors[i] = self._copy_color_in(col, None)
        self.color_changed(i)

    ## Color name access

    def get_color_name(self, i):
        """Looks up a color's name by its list index."""
        try:
            col = self._colors[i]
        except IndexError:
            return
        if col is self._EMPTY_SLOT_ITEM:
            return
        return col.__name

    def set_color_name(self, i, name):
        """Sets a color's name by its list index."""
        try:
            col = self._colors[i]
        except IndexError:
            return
        if col is self._EMPTY_SLOT_ITEM:
            return
        col.__name = name
        self.color_changed(i)

    def get_color_by_name(self, name):
        """Looks up the first color with the given name.

          >>> pltt = Palette()
          >>> pltt.append(RGBColor(1,0,1), "Magenta")
          >>> pltt.get_color_by_name("Magenta")
          <RGBColor r=1.0000, g=0.0000, b=1.0000>

        """
        for col in self:
            if col.__name == name:
                return RGBColor(color=col)

    def __iter__(self):
        return self.iter_colors()

    def iter_colors(self):
        """Iterates across the palette's colors."""
        for col in self._colors:
            if col is self._EMPTY_SLOT_ITEM:
                yield None
            else:
                yield col

    ## Observable events

    @event
    def info_changed(self):
        """Event: palette name, or number of columns was changed."""

    @event
    def match_changed(self):
        """Event: either match position or match_is_approx was updated."""

    @event
    def sequence_changed(self):
        """Event: the color ordering or palette length was changed."""

    @event
    def color_changed(self, i):
        """Event: the color in the given slot, or its name, was modified."""

    ## Dumping and cloning

    def __unicode__(self):
        """Py2-era serialization as a Unicode string.

        Used by the Py3 __str__() while we are in transition.

        """
        result = u"GIMP Palette\n"
        if self._name is not None:
            result += u"Name: %s\n" % self._name
        if self._columns > 0:
            result += u"Columns: %d\n" % self._columns
        result += u"#\n"
        for col in self._colors:
            if col is self._EMPTY_SLOT_ITEM:
                col_name = self._EMPTY_SLOT_NAME
                r = g = b = 0
            else:
                col_name = col.__name
                r, g, b = [clamp(int(c*0xff), 0, 0xff) for c in col.get_rgb()]
            if col_name is None:
                result += u"%d %d %d\n" % (r, g, b)
            else:
                result += u"%d %d %d    %s\n" % (r, g, b, col_name)
        return result

    def __str__(self):
        """Py3: serialize as str (=Unicode). Py2: as bytes (lossy!)."""
        s = self.__unicode__()
        if not PY3:
            s = s.encode("utf-8", errors="replace")
        return s

    def __copy__(self):
        clone = Palette()
        clone.set_name(self.get_name())
        clone.set_columns(self.get_columns())
        for col in self._colors:
            if col is self._EMPTY_SLOT_ITEM:
                clone.append(None)
            else:
                clone.append(copy(col), col.__name)
        return clone

    def __deepcopy__(self, memo):
        return self.__copy__()

    def __repr__(self):
        return "<Palette colors=%d, columns=%d, name=%r>" % (
            len(self._colors),
            self._columns,
            self._name,
        )

    ## Conversion to/from simple dict representation

    def to_simple_dict(self):
        """Converts the palette to a simple dict form used in the prefs."""
        simple = {}
        simple["name"] = self.get_name()
        simple["columns"] = self.get_columns()
        entries = []
        for col in self.iter_colors():
            if col is None:
                entries.append(None)
            else:
                name = col.__name
                entries.append((col.to_hex_str(), name))
        simple["entries"] = entries
        return simple

    @classmethod
    def new_from_simple_dict(cls, simple):
        """Constructs and returns a palette from the simple dict form."""
        pal = cls()
        pal.set_name(simple.get("name", None))
        pal.set_columns(simple.get("columns", None))
        for entry in simple.get("entries", []):
            if entry is None:
                pal.append(None)
            else:
                s, name = entry
                col = RGBColor.new_from_hex_str(s)
                pal.append(col, name)
        return pal


## Helper functions

def _outwards_from(n, i):
    """Search order within the palette, outwards from a given index.

    Defined for a sequence of len() `n`, outwards from index `i`.
    """
    assert i < n and i >= 0
    yield i
    for j in xrange(n):
        exhausted = True
        if i - j >= 0:
            yield i - j
            exhausted = False
        if i + j < n:
            yield i + j
            exhausted = False
        if exhausted:
            break


def _color_distance(c1, c2):
    """Distance metric for color matching in the palette.

    Use a geometric YCbCr distance, as recommended by Graphics Programming with
    Perl, chapter 1, Martien Verbruggen. If we want to give the chrominance
    dimensions a different weighting to luma, we can.

    """
    c1 = YCbCrColor(color=c1)
    c2 = YCbCrColor(color=c2)
    d_cb = c1.Cb - c2.Cb
    d_cr = c1.Cr - c2.Cr
    d_y = c1.Y - c2.Y
    return ((d_cb**2) + (d_cr**2) + (d_y)**2) ** (1.0/3)


## Module testing


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    import doctest
    doctest.testmod()



# This file is part of MyPaint.
# Copyright (C) 2010-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Base classes for window types"""


## Imports

from __future__ import division, print_function
import logging

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GLib

from lib.helpers import clamp, Rect
from lib.pycompat import xrange

logger = logging.getLogger(__name__)


## Base class definitions

class Dialog (Gtk.Dialog):
    """Base dialog accepting all keyboard input.

    Dialogs hide when closed. By default, they accept all keyboard input and
    are not modal. They can (and should) be kept around as references, and can
    be freely hidden and shown after construction.

    """
    def __init__(self, app, *args, **kwargs):
        Gtk.Dialog.__init__(self, *args, **kwargs)
        self.app = app
        if app and app.drawWindow:
            self.set_transient_for(app.drawWindow)
        self.connect('delete-event', lambda w, e: self.hide_on_delete())


class SubWindow (Gtk.Window):
    """A subwindow in the GUI.

    SubWindows don't accept keyboard input by default, but if your subclass
    requires it, pass key_input to the constructor.

    """

    def __init__(self, app, key_input=False):
        """Initialize, as an application subwindow.

        :param app: The main application instance. May be None for testing.
        :param key_input: set to True to accept keyboard input.
        :param hide_on_delete: set False to turn off hide-on-delete behavior.

        """
        Gtk.Window.__init__(self)
        self.app = app
        if app and not key_input:
            self.app.kbm.add_window(self)
            # TODO: do we need a separate class for keyboard-input-friendly
            # windows? Do they share anything in common with dialogs (could
            # they all be implemented as dialogs?)
        self.set_accept_focus(key_input)
        self.pre_hide_pos = None
        # Only hide when the close button is pressed if running as a subwindow
        if app:
            self.connect('delete-event', lambda w, e: self.hide_on_delete())
        # Mark subwindows as utility windows: many X11 WMs handle this sanely
        # This has caused issues with OSX and X11.app under GTK2/PyGTK in the
        # past. OSX builds no longer use X11.app, so this should no longer
        # need special-casing. Testers: use if not sys.platform == 'darwin':
        # if needed, and please submit a patch. https://gna.org/bugs/?15838
        self.set_type_hint(Gdk.WindowTypeHint.UTILITY)
        # Win32 and some Linux DEs are responsive to the following: keeps the
        # window above the main window in fullscreen.
        if app:
            self.set_transient_for(app.drawWindow)

    def show_all(self):
        pos = self.pre_hide_pos
        Gtk.Window.show_all(self)
        if pos:
            self.move(*pos)
        # To keep Compiz happy, move() must be called in the very same
        # handler as show_all(), immediately after. Wiring it in to
        # a map-event or show event handler won't do. Workaround for
        # https://bugs.launchpad.net/ubuntu/+source/compiz/+bug/155101

    def hide(self):
        self.pre_hide_pos = self.get_position()
        Gtk.Window.hide(self)


class PopupWindow (Gtk.Window):
    """
    A popup window, with no decoration. Popups always appear centred under the
    mouse, and don't accept keyboard input.
    """
    def __init__(self, app):
        Gtk.Window.__init__(self, type=Gtk.WindowType.POPUP)
        self.set_gravity(Gdk.Gravity.CENTER)
        self.set_position(Gtk.WindowPosition.MOUSE)
        self.app = app
        self.app.kbm.add_window(self)


class ChooserPopup (Gtk.Window):
    """A resizable popup window used for making fast choices

    Chooser popups can be used for fast selection of items
    from a list of alternatives.
    They normally appear under the mouse pointer,
    but can be popped up next to another on-screen widget
    to provide a menu-like response.

    The popup can be resized using its edges.
    To cancel and hide the popup without making a choice,
    move the pointer outside the window beyond a certain distance or
    click outside the window.

    Code using this class should also hide() the popup
    when the user has made a definite, complete choice
    from what's on offer.

    Popup choosers theoretically permit keyboard input as far as the WM
    is concerned, but eat most keypresses except those whose actions
    have been nominated to be dispatched via ``app.kbm``. As such,
    they're not suited for keyboard data entry, but are fine for
    clicking on brushes, colours etc.

    """

    ## Class constants

    MIN_WIDTH = 256
    MIN_HEIGHT = 256
    MAX_HEIGHT = 512
    MAX_WIDTH = 512

    LEAVE_SLACK = 64
    EDGE_SIZE = 12
    EDGE_CURSORS = {
        None: None,
        Gdk.WindowEdge.NORTH_EAST: Gdk.CursorType.TOP_RIGHT_CORNER,
        Gdk.WindowEdge.NORTH_WEST: Gdk.CursorType.TOP_LEFT_CORNER,
        Gdk.WindowEdge.SOUTH_EAST: Gdk.CursorType.BOTTOM_RIGHT_CORNER,
        Gdk.WindowEdge.SOUTH_WEST: Gdk.CursorType.BOTTOM_LEFT_CORNER,
        Gdk.WindowEdge.WEST: Gdk.CursorType.LEFT_SIDE,
        Gdk.WindowEdge.EAST: Gdk.CursorType.RIGHT_SIDE,
        Gdk.WindowEdge.SOUTH: Gdk.CursorType.BOTTOM_SIDE,
        Gdk.WindowEdge.NORTH: Gdk.CursorType.TOP_SIDE,
    }

    ## Method defs

    def __init__(self, app, actions, config_name):
        """Initialize.

        :param app: the main Application object.
        :param iterable actions: keyboard action names to pass through.
        :param str config_name: config prefix for saving window size.

        Use a simple "lowercase_with_underscores" name for the
        configuration key prefix.

        See also: `gui.keyboard.KeyboardManager.add_window()`.
        """
        # Superclass
        Gtk.Window.__init__(self, type=Gtk.WindowType.POPUP)
        self.set_modal(True)

        # Internal state
        self.app = app
        self._size = None  # last recorded size from any show()
        self._motion_handler_id = None
        self._prefs_size_key = "%s.window_size" % (config_name,)
        self._resize_info = None   # state during an edge resize
        self._outside_grab_active = False
        self._outside_cursor = Gdk.Cursor(Gdk.CursorType.LEFT_PTR)
        self._popup_info = None

        # Initial positioning
        self._initial_move_pos = None  # used when forcing a specific position
        self._corrected_pos = None  # used when keeping the widget on-screen

        # Resize cursors
        self._edge_cursors = {}
        for edge, cursor in self.EDGE_CURSORS.items():
            if cursor is not None:
                cursor = Gdk.Cursor(cursor)
            self._edge_cursors[edge] = cursor

        # Default size
        self.set_gravity(Gdk.Gravity.NORTH_WEST)
        default_size = (self.MIN_WIDTH, self.MIN_HEIGHT)
        w, h = app.preferences.get(self._prefs_size_key, default_size)
        w = clamp(int(w), self.MIN_WIDTH, self.MAX_WIDTH)
        h = clamp(int(h), self.MIN_HEIGHT, self.MAX_HEIGHT)
        default_size = (w, h)
        self.set_transient_for(app.drawWindow)
        self.set_default_size(*default_size)
        self.set_position(Gtk.WindowPosition.MOUSE)

        # Register with the keyboard manager, but only let certain actions be
        # driven from the keyboard.
        app.kbm.add_window(self, actions)

        # Event handlers
        self.connect("realize", self._realize_cb)
        self.connect("configure-event", self._configure_cb)
        self.connect("enter-notify-event", self._crossing_cb)
        self.connect("leave-notify-event", self._crossing_cb)
        self.connect("show", self._show_cb)
        self.connect("hide", self._hide_cb)
        self.connect("button-press-event", self._button_press_cb)
        self.connect("button-release-event", self._button_release_cb)
        self.add_events(Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK)

        # Appearance
        self._frame = Gtk.Frame()
        self._frame.set_shadow_type(Gtk.ShadowType.OUT)
        self._align = Gtk.Alignment.new(0.5, 0.5, 1.0, 1.0)
        self._align.set_padding(self.EDGE_SIZE, self.EDGE_SIZE,
                                self.EDGE_SIZE, self.EDGE_SIZE)
        self._frame.add(self._align)
        Gtk.Window.add(self, self._frame)

    def _crossing_cb(self, widget, event):
        if self._resize_info:
            return
        if event.mode != Gdk.CrossingMode.NORMAL:
            return
        if event.detail != Gdk.NotifyType.NONLINEAR:
            return
        if event.get_window() is not self.get_window():
            return
        x, y, w, h = self._get_size()
        inside = (x <= event.x_root < x+w) and (y <= event.y_root < y+h)
        logger.debug("crossing: inside=%r", inside)
        # Grab the pointer if crossing from inside the popup to the
        # outside. Ungrab if doing the reverse.
        if inside:
            self._ungrab_pointer_outside(
                device = event.get_device(),
                time = event.time,
            )
        else:
            self._grab_pointer_outside(
                device = event.get_device(),
                time = event.time,
            )

    def _grab_pointer_outside(self, device, time):
        if self._outside_grab_active:
            logger.warning("grab: outside-popup grab already active: "
                           "regrabbing")
            self._ungrab_pointer_outside(device, time)
        event_mask = (
            Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.ENTER_NOTIFY_MASK
            | Gdk.EventMask.LEAVE_NOTIFY_MASK
            | Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
        )
        cursor = self._outside_cursor
        grab_status = device.grab(
            window = self.get_window(),
            grab_ownership = Gdk.GrabOwnership.APPLICATION,
            owner_events = False,
            event_mask = Gdk.EventMask(event_mask),
            cursor = cursor,
            time_ = time,
        )
        if grab_status == Gdk.GrabStatus.SUCCESS:
            logger.debug("grab: acquired grab on %r successfully", device)
            self._outside_grab_active = True
        else:
            logger.warning("grab: failed to acquire grab on %r (status=%s)",
                           device, grab_status.value_nick)

    def _ungrab_pointer_outside(self, device, time):
        if not self._outside_grab_active:
            logger.debug("ungrab: outside-popup grab not active")
        device.ungrab(time_=time)
        logger.debug("ungrab: released grab on %r", device)
        self._outside_grab_active = False

    def _configure_cb(self, widget, event):
        """Internal: Update size and prefs when window is adjusted"""
        # Constrain window to fit on its current monitor, if possible.
        screen = event.get_screen()
        mon = screen.get_monitor_at_point(event.x, event.y)
        mon_geom = screen.get_monitor_geometry(mon)
        # Constrain width and height
        w = clamp(int(event.width), self.MIN_WIDTH, self.MAX_WIDTH)
        h = clamp(int(event.height), self.MIN_HEIGHT, self.MAX_HEIGHT)
        # Constrain position
        x, y = event.x, event.y
        if y+h > mon_geom.y + mon_geom.height:
            y = mon_geom.y + mon_geom.height - h
        if x+w > mon_geom.x + mon_geom.width:
            x = mon_geom.x + mon_geom.width - w
        if x < mon_geom.x:
            x = mon_geom.x
        if y < mon_geom.y:
            y = mon_geom.y
        event_size = (event.x, event.y, event.width, event.height)
        ex, ey, ew, eh = [int(c) for c in event_size]
        x, y, w, h = [int(c) for c in (x, y, w, h)]
        if not self._corrected_pos:
            if (x, y) != (ex, ey):
                GLib.idle_add(self.move, x, y)
            if (w, h) != (ew, eh):
                GLib.idle_add(self.resize, w, h)
            self._corrected_pos = True
        # Record size
        self._size = (x, y, w, h)
        self.app.preferences[self._prefs_size_key] = (w, h)

    def _get_size(self):
        if not self._size:
            # From time to time, popups presented in fullscreen don't
            # receive configure events. Why?
            win = self.get_window()
            x, y = win.get_position()
            w = win.get_width()
            h = win.get_height()
            x, y = self.get_position()
            self._size = (x, y, w, h)
        return self._size

    def _realize_cb(self, widget):
        gdk_window = self.get_window()
        gdk_window.set_type_hint(Gdk.WindowTypeHint.POPUP_MENU)

    def popup(self, widget=None, above=False, textwards=True, event=None):
        """Display, with an optional position relative to a widget

        :param widget: The widget defining the pop-up position
        :param above: If true, pop up above from `widget`
        :param textwards: If true, pop up in the text direction from `widget`
        :param event: the originating event

        """
        if not widget:
            self.set_position(Gtk.WindowPosition.MOUSE)
            self.set_gravity(Gdk.Gravity.NORTH_WEST)
        else:
            win = widget.get_window()
            x, y = win.get_origin()[1:]
            alloc = widget.get_allocation()
            style = widget.get_style_context()
            rtl = (style.get_direction() == Gtk.TextDirection.RTL)
            grav_table = {
                # (Above, rtl, textwards): Gravity
                (True, True, True): Gdk.Gravity.SOUTH_EAST,
                (True, False, False): Gdk.Gravity.SOUTH_EAST,
                (True, True, False): Gdk.Gravity.SOUTH_WEST,
                (True, False, True): Gdk.Gravity.SOUTH_WEST,
                (False, True, True): Gdk.Gravity.NORTH_EAST,
                (False, False, False): Gdk.Gravity.NORTH_EAST,
                (False, True, False): Gdk.Gravity.NORTH_WEST,
                (False, False, True): Gdk.Gravity.NORTH_WEST,
            }
            grav = grav_table.get((above, rtl, textwards),
                                  Gdk.Gravity.NORTH_WEST)
            if not above:
                y += alloc.height
            if (rtl and textwards) or ((not rtl) and (not textwards)):
                x += alloc.width
            self.set_position(Gtk.WindowPosition.NONE)
            self.set_gravity(grav)
            x = int(x)
            y = int(y)
            self._initial_move_pos = (x, y)
            if self._size:
                self._do_initial_move()
        popup_info = None
        if event:
            popup_info = (event.get_device(), event.time)
        self._popup_info = popup_info
        self.present()

    def _do_initial_move(self):
        x, y, w, h = self._get_size()
        x, y = self._initial_move_pos
        grav = self.get_gravity()
        if grav in [Gdk.Gravity.SOUTH_EAST, Gdk.Gravity.SOUTH_WEST]:
            y -= h
        if grav in [Gdk.Gravity.NORTH_EAST, Gdk.Gravity.SOUTH_EAST]:
            x -= w
        self.move(x, y)

    def _show_cb(self, widget):
        """Internal: show child widgets, grab, start the motion handler"""
        self._frame.show_all()
        if not self._motion_handler_id:
            h_id = self.connect("motion-notify-event", self._motion_cb)
            self._motion_handler_id = h_id
        if self._initial_move_pos:
            self._do_initial_move()
            self._initial_move_pos = None
            self.set_gravity(Gdk.Gravity.NORTH_WEST)
        # Grab the device used for the click
        # if popped up next to a widget - the cursor counts as outside
        # initially.
        if self._popup_info:
            device, time = self._popup_info
            self._grab_pointer_outside(
                device = device,
                time = time,
            )

    def _hide_cb(self, widget):
        """Internal: reset during-show state when the window is hidden"""
        if self._motion_handler_id is not None:
            self.disconnect(self._motion_handler_id)
        self._motion_handler_id = None
        self._corrected_pos = None
        self._initial_move_pos = None
        self._popup_info = None
        self._resize_info = None
        self._outside_grab_active = False

    def add(self, child):
        """Override: add() adds the child widget to an internal alignment"""
        return self._align.add(child)

    def remove(self, child):
        """Override: remove() removes the child from an internal alignment"""
        return self._align.remove(child)

    def _get_edge(self, px, py):
        """Internal: returns which window edge the pointer is pointing at"""
        size = self._get_size()
        if size is None:
            return None
        x, y, w, h = size
        s = self.EDGE_SIZE
        inside = (x <= px < x+w) and (y <= py < y+h)
        if not inside:
            return None
        north = east = south = west = False
        if px >= x and px <= x+s:
            west = True
        elif px <= x+w and px >= x+w-s:
            east = True
        if py >= y and py <= y+s:
            north = True
        elif py <= y+h and py >= y+h-s:
            south = True
        if north:
            if east:
                return Gdk.WindowEdge.NORTH_EAST
            elif west:
                return Gdk.WindowEdge.NORTH_WEST
            else:
                return Gdk.WindowEdge.NORTH
        elif south:
            if east:
                return Gdk.WindowEdge.SOUTH_EAST
            elif west:
                return Gdk.WindowEdge.SOUTH_WEST
            else:
                return Gdk.WindowEdge.SOUTH
        elif east:
            return Gdk.WindowEdge.EAST
        elif west:
            return Gdk.WindowEdge.WEST
        else:
            return None

    def _get_cursor(self, px, py):
        x, y, w, h = self._get_size()
        edge = self._get_edge(px, py)
        outside_window = px < x or py < y or px > x+w or py > y+h
        if outside_window:
            return self._outside_cursor
        else:
            return self._edge_cursors.get(edge, None)

    def _button_press_cb(self, widget, event):
        """Internal: starts resizing if the pointer is at the edge"""
        win = self.get_window()
        if not win:
            return False
        if event.button != 1:
            return False
        if not self.get_visible():
            return False

        rx, ry = event.x_root, event.y_root
        edge = self._get_edge(rx, ry)
        size = self._get_size()
        if edge is not None:
            self._resize_info = (rx, ry, size, edge)
            return True
        else:
            x, y, w, h = size
            inside = (x <= rx < x+w) and (y <= ry < y+h)
            if not inside:
                logger.debug("click outside detected: hiding popup")
                self.hide()
                return True
        return False

    def _button_release_cb(self, widget, event):
        """Internal: stops any active resize"""
        if event.button != 1:
            return False
        if not self.get_visible():
            return False
        if self._resize_info:
            self._resize_info = None
        # IDEA: re-grab the pointer here if the event is inside?
        return True

    def _motion_cb(self, widget, event):
        """Internal: handle motions: resizing, or leave checks"""

        # Ensure that at some point the user started inside a visible window
        if not self.get_visible():
            return
        win = self.get_window()
        if not win:
            return
        size = self._get_size()
        assert size is not None

        # Resizing
        px, py = event.x_root, event.y_root
        if self._resize_info:
            px0, py0, size0, edge0 = self._resize_info
            x0, y0, w0, h0 = size0
            dx, dy = (px-px0, py-py0)
            x, y = x0, y0
            w, h = w0, h0
            if edge0 in (Gdk.WindowEdge.NORTH_WEST, Gdk.WindowEdge.WEST,
                         Gdk.WindowEdge.SOUTH_WEST):
                x += dx
                w -= dx
            elif edge0 in (Gdk.WindowEdge.NORTH_EAST, Gdk.WindowEdge.EAST,
                           Gdk.WindowEdge.SOUTH_EAST):
                w += dx
            if edge0 in (Gdk.WindowEdge.NORTH_WEST, Gdk.WindowEdge.NORTH,
                         Gdk.WindowEdge.NORTH_EAST):
                y += dy
                h -= dy
            elif edge0 in (Gdk.WindowEdge.SOUTH_EAST, Gdk.WindowEdge.SOUTH,
                           Gdk.WindowEdge.SOUTH_WEST):
                h += dy

            # Apply constraints
            # One caveat: if the minimum size of the widgets themselves
            # is larger than the hardcoded MIN_* sizes here, moving the
            # top or left edges can look weird until the calculated w or h
            # butts up against the minimum: it effectively does a move.
            if w < self.MIN_WIDTH:
                if x != x0:
                    x -= self.MIN_WIDTH - w
                w = self.MIN_WIDTH
            if h < self.MIN_HEIGHT:
                if y != y0:
                    y -= self.MIN_HEIGHT - h
                h = self.MIN_HEIGHT

            # Move and/or resize
            # Can't use gdk_window_move_resize because that screws up
            # child widgets with GtkScrolledWindows.
            if (x, y) != (x0, y0):
                self.move(x, y)
            if (w, h) != (w0, h0):
                self.resize(w, h)

            return True

        x, y, w, h = size
        inside = (x <= px < x+w) and (y < py < y+h)

        # One oddity of the outside grab handling code is that it
        # doesn't always detect re-entry (in fact, it's cyclic; every
        # *other* crossing to the inside gets dropped). As a workaround,
        # if the grab is active and the cursor is *moving* inside,
        # release the grab so that the user can interact with the
        # window's contents.
        if inside and self._outside_grab_active:
            self._ungrab_pointer_outside(event.get_device(), event.time)

        cursor = self._get_cursor(px, py)
        win.set_cursor(cursor)

        if inside:
            return False

        if self._popup_info:
            return False

        any_button_mask = (
            Gdk.ModifierType.BUTTON1_MASK
            | Gdk.ModifierType.BUTTON2_MASK
            | Gdk.ModifierType.BUTTON3_MASK
            | Gdk.ModifierType.BUTTON4_MASK
            | Gdk.ModifierType.BUTTON5_MASK
        )

        if event.state & any_button_mask:
            return False

        # Moving outside the window more than a certain amount causes it
        # to close.
        s = self.LEAVE_SLACK
        outside_tolerance = px < x-s or py < y-s or px > x+w+s or py > y+h+s
        if outside_tolerance:
            self.hide()
            return True


# Window positioning helper functions

def _final_rectangle(
        x, y, w, h, screen_w, screen_h, targ_geom, min_usable_size):
    """ Tries to create sensible (x, y, w, h) window pos/dim
    """
    # Generate a sensible, positive x and y position
    final_x, final_y, final_w, final_h = None, None, None, None
    if x is not None and y is not None:
        if x >= 0:
            final_x = x
        else:
            assert w is not None
            assert w > 0
            final_x = targ_geom.x + (targ_geom.w - w - abs(x))
        if y >= 0:
            final_y = y
        else:
            assert h is not None
            assert h > 0
            final_y = targ_geom.y + (targ_geom.h - h - abs(y))
        if final_x < 0 or final_x > screen_w - min_usable_size:
            final_x = None
        if final_y < 0 or final_y > screen_h - min_usable_size:
            final_y = None
    # and a sensible, positive width and height
    if w is not None and h is not None:
        final_w = w
        final_h = h
        if w < 0 or h < 0:
            if w < 0:
                if x is not None:
                    final_w = max(0, targ_geom.w - abs(x) - abs(w))
                else:
                    final_w = max(0, targ_geom.w - 2*abs(w))
            if h < 0:
                if x is not None:
                    final_h = max(0, targ_geom.h - abs(y) - abs(h))
                else:
                    final_h = max(0, targ_geom.h - 2*abs(h))
        if final_w > screen_w or final_w < min_usable_size:
            final_w = None
        if final_h > screen_h or final_h < min_usable_size:
            final_h = None

    return final_x, final_y, final_w, final_h


def set_initial_window_position(win, pos):
    """Set the position of a Gtk.Window, used during initial positioning.

    This is used both for restoring a saved window position, and for the
    application-wide defaults. The ``pos`` argument is a dict containing the
    following optional keys

        "w": <int>
        "h": <int>
            If positive, the size of the window.
            If negative, size is calculated based on the size of the
            monitor with the pointer on it, and x (or y) if given, e.g.

                width = mouse_mon_w -  abs(x) + abs(w)   # or (if no x)
                width = mouse_mon_w - (2 * abs(w))

            The same is true of calculated heights.

        "x": <int>
        "y": <int>
            If positive, the left/top of the window.
            If negative, the bottom/right of the window on the monitor
            with the pointer on it: you MUST provide a positive w and h
            if you do this.

    If the window's calculated top-left would place it offscreen, it will be
    placed in its default, window manager provided position. If its calculated
    size is larger than the screen, the window will be given its natural size
    instead.

    Returns the final, chosen (x, y) pair for forcing the window position on
    first map, or None if defaults are being used.

    """

    min_usable_size = 100

    # Where the mouse is right now - identifies the current monitor.
    ptr_x, ptr_y = 0, 0
    screen = win.get_screen()
    display = win.get_display()
    devmgr = display and display.get_device_manager() or None
    ptrdev = devmgr and devmgr.get_client_pointer() or None
    if ptrdev:
        ptr_screen, ptr_x, ptr_y = ptrdev.get_position()
        assert ptr_screen is screen, (
            "Screen containing core pointer != screen containing "
            "the window for positioning (%r != %r)" % (ptr_screen, screen)
        )
        logger.debug("Core pointer position from display: %r", (ptr_x, ptr_y))
    else:
        logger.warning(
            "Could not determine core pointer position from display. "
            "Using %r instead.",
            (ptr_x, ptr_y),
        )
    screen_w = screen.get_width()
    screen_h = screen.get_height()
    assert screen_w > min_usable_size
    assert screen_h > min_usable_size

    # The target area is ideally the current monitor.
    targ_mon_num = screen.get_monitor_at_point(ptr_x, ptr_y)
    targ_geom = _get_target_area_geometry(screen, targ_mon_num)

    # Positioning arguments
    x, y, w, h = (pos.get(k, None) for k in ("x", "y", "w", "h"))
    final_x, final_y, final_w, final_h = _final_rectangle(
        x, y, w, h, screen_w, screen_h, targ_geom, min_usable_size)

    # If the window is positioned, make sure it's on a monitor which still
    # exists. Users change display layouts...
    if None not in (final_x, final_y):
        onscreen = False
        for mon_num in xrange(screen.get_n_monitors()):
            targ_geom = _get_target_area_geometry(screen, mon_num)
            in_targ_geom = (
                final_x < (targ_geom.x + targ_geom.w)
                and final_y < (targ_geom.x + targ_geom.h)
                and final_x >= targ_geom.x
                and final_y >= targ_geom.y
            )
            if in_targ_geom:
                onscreen = True
                break
        if not onscreen:
            logger.warning("Calculated window position is offscreen; "
                           "ignoring %r" % ((final_x, final_y), ))
            final_x = None
            final_y = None

    # Attempt to set up with a geometry string first. Repeats the block below
    # really, but this helps smaller windows receive the right position in
    # xfwm (at least), possibly because the right window hints will be set.
    if None not in (final_w, final_h, final_x, final_y):
        geom_str = "%dx%d+%d+%d" % (final_w, final_h, final_x, final_y)
        win.connect("realize", lambda *a: win.parse_geometry(geom_str))

    # Set what we can now.
    if None not in (final_w, final_h):
        win.set_default_size(final_w, final_h)
    if None not in (final_x, final_y):
        win.move(final_x, final_y)
        return final_x, final_y

    return None


def _get_target_area_geometry(screen, mon_num):
    """Get a rect for putting windows in: normally based on monitor.

    :param Gdk.Screen screen: Target screen.
    :param int mon_num: Monitor number, e.g. that of the pointer.
    :returns: A hopefully usable target area.
    :rtype: Rect

    This function operates like gdk_screen_get_monitor_geometry(), but
    falls back to the screen geometry for cases when that returns NULL.
    It also returns a type which has (around GTK 3.18.x) fewer weird
    typelib issues with construction or use.

    Ref: https://github.com/mypaint/mypaint/issues/424
    Ref: https://github.com/mypaint/mypaint/issues/437

    """
    geom = None
    if mon_num >= 0:
        geom = screen.get_monitor_geometry(mon_num)
    if geom is not None:
        geom = Rect.new_from_gdk_rectangle(geom)
    else:
        logger.warning(
            "gdk_screen_get_monitor_geometry() returned NULL: "
            "using screen size instead as a fallback."
        )
        geom = Rect(0, 0, screen.get_width(), screen.get_height())
    return geom
